import os
import uuid
from datetime import datetime, timezone
import time
import numpy as np
import pandas as pd

from hystore.core import operations
from hystore.core import persistence as per
from hystore.core import fields as fld
from hystore.core import readerwriter as rw
from hystore.core import validation as val
from hystore.core import operations as ops
from hystore.core import utils


# TODO:
"""
 * joins: get mapping and map multiple fields using mapping
   * can use pandas for initial functionality and then improve
   * mark fields has having sorted order if they are sorted
     * would have to check field
   * for sorted, can use fast
   * aggregation
     * aggregate and join / join and aggregate
       * if they resolve to being the same thing, best to aggregate first
 * everything can accept groups
 * groups are registered and given names
 * indices can be built from merging pk and fks and mapping to values
 * everything can accept tuples instead of groups and operate on all of them
 * advanced sorting / filtering
   * FilteredReader / FilteredWriter for applying filters without copying data
   * SortedReader / SortedWriter for applying sorts without copying data
   * Filters corresponding to the presence or absence of elements of a maybe field
     * should be marked indicating that they are related to that field. This allows
       a user to query all the filters generated for different fields, whether the
       values are valid or whether they are a standard filtering for different
       categories (age sub-ranges, for example)
 * reader/writers
   * should be able to read/write through the same object (they can now but not like this)
     rw = s.get(name)
     rw.writeable.data[:] = data
     data = rw.data[:]
   * soft sorting / filtering
     rw = s.get(name)
     rw.add_transform(filter)
     rw.add_transform(sort)
     rw.writeable[:] = data # can this be done?
     rw.clean()
     rw.write(data)
"""

class Session:

    def __init__(self, chunksize=ops.DEFAULT_CHUNKSIZE,
                 timestamp=str(datetime.now(timezone.utc))):
        if not isinstance(timestamp, str):
            error_str = "'timestamp' must be a string but is of type {}"
            raise ValueError(error_str.format(type(timestamp)))
        self.chunksize = chunksize
        self.timestamp = timestamp


    def get_shared_index(self, keys):
        """
        Create a shared index based on a tuple numpy arrays containing keys.
        This function generates the sorted union of a tuple of key fields and
        then maps the individual arrays to their corresponding indices in the
        sorted union.

        Example:
            key_1 = ['a', 'b', 'e', 'g', 'i']
            key_2 = ['b', 'b', 'c', 'c, 'e', 'g', 'j']
            key_3 = ['a', 'c' 'd', 'e', 'g', 'h', 'h', 'i']

            sorted_union = ['a', 'b', 'c', 'd', 'e', 'g', 'h', 'i', 'j']

            key_1_index = [0, 1, 4, 5, 7]
            key_2_index = [1, 1, 2, 2, 4, 5, 8]
            key_3_index = [0, 2, 3, 4, 5, 6, 6, 7]

            :param keys: a tuple of groups, fields or ndarrays whose contents represent keys
        """
        if not isinstance(keys, tuple):
            raise ValueError("'keys' must be a tuple")
        concatted = None
        raw_keys = list()
        for k in keys:
            raw_field = val.raw_array_from_parameter(self, 'keys', k)
            raw_keys.append(raw_field)
            if concatted is None:
                concatted = pd.unique(raw_field)
            else:
                concatted = np.concatenate((concatted, raw_field), axis=0)
        concatted = pd.unique(concatted)
        concatted = np.sort(concatted)

        return tuple(np.searchsorted(concatted, k) for k in raw_keys)


    def set_timestamp(self, timestamp=str(datetime.now(timezone.utc))):
        """
        Set the default timestamp to be used when creating fields without an explicit
        timestamp specified.

        :param timestamp a string representing a valid Datetime
        :return: None
        """
        if not isinstance(timestamp, str):
            error_str = "'timestamp' must be a string but is of type {}"
            raise ValueError(error_str.format(type(timestamp)))
        self.timestamp = timestamp



    def sort_on(self, src_group, dest_group, keys,
                timestamp=datetime.now(timezone.utc), write_mode='write'):
        """
        Sort a group (src_group) of fields by the specified set of keys, and write the
        sorted fields to dest_group.

        :param src_group: the group of fields that are to be sorted
        :param dest_group: the group into which sorted fields are written
        :param keys: fields to sort on
        :param timestamp: optional - timestamp to write on the sorted fields
        :param write_mode: optional - write mode to use if the destination fields already
        exist
        :return: None
        """
        # TODO: fields is being ignored at present

        readers = tuple(self.get_reader(src_group[f]) for f in keys)
        t1 = time.time()
        sorted_index = self.dataset_sort_index(
            np.arange(len(readers[0]), dtype=np.uint32), readers)
        print(f'sorted {keys} index in {time.time() - t1}s')

        t0 = time.time()
        for k in src_group.keys():
            t1 = time.time()
            r = self.get_reader(src_group[k])
            w = r.get_writer(dest_group, k, timestamp, write_mode=write_mode)
            self.apply_index(sorted_index, r, w)
            del r
            del w
            print(f"  '{k}' reordered in {time.time() - t1}s")
        print(f"fields reordered in {time.time() - t0}s")



    def dataset_sort_index(self, sort_indices, index=None):
        """
        Generate a sorted index based on a set of fields upon which to sort and an optional
        index to apply to the sort_indices
        :param sort_indices: a tuple or list of indices that determine the sorted order
        :param index: optional - the index by which the initial field should be permuted
        :return: the resulting index that can be used to permute unsorted fields
        """
        val._check_all_readers_valid_and_same_type(sort_indices)
        r_readers = tuple(reversed(sort_indices))

        raw_data = val.raw_array_from_parameter(self, 'readers', r_readers[0])

        if index is None:
            raw_index = np.arange(len(raw_data))
        else:
            raw_index = val.raw_array_from_parameter(self, 'index', index)

        acc_index = raw_index
        fdata = raw_data[acc_index]
        index = np.argsort(fdata, kind='stable')
        acc_index = acc_index[index]

        for r in r_readers[1:]:
            raw_data = val.raw_array_from_parameter(self, 'readers', r)
            fdata = raw_data[acc_index]
            index = np.argsort(fdata, kind='stable')
            acc_index = acc_index[index]

        return acc_index


    def apply_filter(self, filter_to_apply, src, dest=None):
        """
        Apply a filter to an a src field. The filtered field is written to dest if it set,
        and returned from the function call. If the field is an IndexedStringField, the
        indices and values are returned separately.

        :param filter_to_apply: the filter to be applied to the source field
        :param src: the field to be filtered
        :param dest: optional - a field to write the filtered data to
        :return: the filtered values
        """
        filter_to_apply_ = val.array_from_parameter(self, 'index_to_apply', filter_to_apply)
        writer_ = None
        if writer_:
            writer_ = val.field_from_parameter(self, 'writer', dest)
        if isinstance(src, fld.IndexedStringField):
            src_ = val.field_from_parameter(self, 'reader', src)
            dest_indices, dest_values =\
                ops.apply_filter_to_index_values(filter_to_apply_,
                                                 src_.indices[:], src_.values[:])
            if writer_:
                writer_.write_raw(dest_indices, dest_values)
            return dest_indices, dest_values
        else:
            reader_ = val.array_from_parameter(self, 'reader', src)
            result = reader_[filter_to_apply]
            if writer_:
                writer_.write(result)
            return result


    def apply_index(self, index_to_apply, src, dest=None):
        """
        Apply a index to an a src field. The indexed field is written to dest if it set,
        and returned from the function call. If the field is an IndexedStringField, the
        indices and values are returned separately.

        :param index_to_apply: the index to be applied to the source field
        :param src: the field to be index
        :param dest: optional - a field to write the indexed data to
        :return: the indexed values
        """
        index_to_apply_ = val.array_from_parameter(self, 'index_to_apply', index_to_apply)
        writer_ = None
        if writer_:
            writer_ = val.field_from_parameter(self, 'writer', dest)
        if isinstance(src, fld.IndexedStringField):
            src_ = val.field_from_parameter(self, 'reader', src)
            dest_indices, dest_values =\
                ops.apply_indices_to_index_values(index_to_apply_,
                                                  src_.indices[:], src_.values[:])
            if writer_:
                writer_.write_raw(dest_indices, dest_values)
            return dest_indices, dest_values
        else:
            reader_ = val.array_from_parameter(self, 'reader', src)
            result = reader_[index_to_apply]
            if writer_:
                writer_.write(result)
            return result


    # TODO: write distinct with new readers / writers rather than deleting this
    def distinct(self, field=None, fields=None, filter=None):
        if field is None and fields is None:
            return ValueError("One of 'field' and 'fields' must be set")
        if field is not None and fields is not None:
            return ValueError("Only one of 'field' and 'fields' may be set")

        if field is not None:
            return np.unique(field)

        entries = [(f'{i}', f.dtype) for i, f in enumerate(fields)]
        unified = np.empty_like(fields[0], dtype=np.dtype(entries))
        for i, f in enumerate(fields):
            unified[f'{i}'] = f

        uniques = np.unique(unified)
        results = [uniques[f'{i}'] for i in range(len(fields))]
        return results


    def get_spans(self, field=None, fields=None):
        """
        Calculate a set of spans that indicate contiguous equal values.
        The entries in the result array correspond to the inclusive start and
        exclusive end of the span (the ith span is represented by element i and
        element i+1 of the result array). The last entry of the result array is
        the length of the source field.

        Only one of 'field' or 'fields' may be set. If 'fields' is used and more
        than one field specified, the fields are effectively zipped and the check
        for spans is carried out on each corresponding tuple in the zipped field.

        Example:
            field: [1, 2, 2, 1, 1, 1, 3, 4, 4, 4, 2, 2, 2, 2, 2]
            result: [0, 1, 3, 6, 7, 10, 15]
        """
        if field is None and fields is None:
            raise ValueError("One of 'field' and 'fields' must be set")
        if field is not None and fields is not None:
            raise ValueError("Only one of 'field' and 'fields' may be set")
        raw_field = None
        raw_fields = None
        if field is not None:
            val._check_is_reader_or_ndarray('field', field)
            raw_field = field[:] if isinstance(field, rw.Reader) else field
        else:
            raw_fields = []
            for f in fields:
                val._check_is_reader_or_ndarray('elements of tuple/list fields', f)
                raw_fields.append(f[:] if isinstance(f, rw.Reader) else f)
        return per._get_spans(raw_field, raw_fields)


    def _apply_spans_no_src(self, predicate, spans, dest=None):
        if dest:
            if val.is_field_parameter(self, 'dest', dest):
                dest_ = val.field_from_parameter(self, 'dest', dest)
                results = np.zeros(len(spans) - 1, dtype=dest_.dtype)
                predicate(spans, results)
                dest_.data.write(results)
                return results
            else:
                dest_ = val.array_from_parameter(self, 'dest', dest)
                if len(dest_) != len(spans) - 1:
                    error_msg = ("if 'dest' (length {}) is an ndarray, it must be one element "
                                 "shorter than the length of 'spans' (length{})")
                    raise ValueError(error_msg.format(len(dest_), len(spans)))
                    predicate(spans, dest_)
                return dest_
        else:
            results = np.zeros(len(spans) - 1, dtype='int64')
            predicate(spans, results)
            return results


    def _apply_spans_src(self, predicate, spans, src, dest=None):
        src_ = val.array_from_parameter(src)
        if len(src) != len(spans) - 1:
            error_msg = ("'src' (length {}) must be one element shorter than 'spans' "
                         "(length {})")
            raise ValueError(error_msg.format(len(src_), len(spans)))

        if dest:
            if val.is_field_parameter(self, 'dest', dest):
                dest_ = val.field_from_parameter(self, 'dest', dest)
                results = np.zeros(len(spans) - 1, dtype=dest_.dtype)
                predicate(spans, results)
                dest_.data.write(results)
                return results
            else:
                dest_ = val.array_from_parameter(self, 'dest', dest)
                if len(dest_) != len(spans) - 1:
                    error_msg = ("if 'dest' (length {}) is an ndarray, it must be one element "
                                 "shorter than the length of 'spans' (length{})")
                    raise ValueError(error_msg.format(len(dest_), len(spans)))
                    predicate(spans, dest_)
                return dest_
        else:
            results = np.zeros(len(spans) - 1, dtype='int64')
            predicate(spans, results)
            return results

    def apply_spans_index_of_min(self, spans, src, dest=None):
        return self._apply_spans_src(ops.apply_spans_index_of_min, spans, src, dest)

    def apply_spans_index_of_max(self, spans, src, dest=None):
        return self._apply_spans_src(ops.apply_spans_index_of_max, spans, src, dest)

    def apply_spans_index_of_first(self, spans, dest=None):
        return self._apply_spans_no_src(ops.apply_spans_index_of_first, spans, dest)

    def apply_spans_index_of_last(self, spans, dest=None):
        return self._apply_spans_no_src(ops.apply_spans_index_of_last, spans, dest)

    def apply_spans_count(self, spans, dest=None):
        return self._apply_spans_no_src(ops.apply_spans_count, spans, dest)

    def apply_spans_min(self, spans, src, dest=None):
        return self._apply_spans_src(ops.apply_spans_min, spans, src, dest)

    def apply_spans_max(self, spans, src, dest=None):
        return self._apply_spans_src(ops.apply_spans_max, spans, src, dest)

    def apply_spans_first(self, spans, dest=None):
        return self._apply_spans_no_src(ops.apply_spans_first, spans, dest)

    def apply_spans_last(self, spans, dest=None):
        return self._apply_spans_no_src(ops.apply_spans_last, spans, dest)

    def apply_spans_concat(self, spans, src, dest):
        if not isinstance(src, fld.IndexedStringField):
            raise ValueError(f"'src' must be one of 'IndexedStringField' but is {type(src)}")
        if not isinstance(dest, fld.IndexedStringField):
            raise ValueError(f"'dest' must be one of 'IndexedStringField' but is {type(dest)}")

        src_index = src.indices[:]
        src_values = src.values[:]
        dest_index = np.zeros(src.chunksize, src_index.dtype)
        dest_values = np.zeros(dest.chunksize * 16, src_values.dtype)

        max_index_i = src.chunksize
        max_value_i = dest.chunksize * 8
        s = 0
        while s < len(spans) - 1:
            s, index_i, index_v = per._apply_spans_concat(spans, src_index, src_values,
                                                          dest_index, dest_values,
                                                          max_index_i, max_value_i, s)

            if index_i > 0 or index_v > 0:
                dest.write_raw(dest_index[:index_i], dest_values[:index_v])
        dest.complete()


    def _aggregate_impl(self, predicate, fkey_indices=None, fkey_index_spans=None,
                        src=None, dest=None, result_dtype=None):
        if fkey_indices is None and fkey_index_spans is None:
            raise ValueError("One of 'fkey_indices' or 'fkey_index_spans' must be set")
        if fkey_indices is not None and fkey_index_spans is not None:
            raise ValueError("Only one of 'fkey_indices' and 'fkey_index_spans' may be set")
        if dest is not None:
            if not isinstance(dest, (fld.Field, np.ndarray)):
                raise ValueError("'writer' must be either a Writer or an ndarray instance")

        if fkey_index_spans is None:
            fkey_index_spans = self.get_spans(field=fkey_indices)

        if isinstance(dest, np.ndarray):
            if len(dest) != len(fkey_index_spans) - 1:
                error = "'writer': ndarray must be of length {} but is of length {}"
                raise ValueError(error.format(len(fkey_index_spans) - 1), len(dest))
            elif dest.dtype != result_dtype:
                raise ValueError(f"'writer' dtype must be {result_dtype} but is {dest.dtype}")

        if isinstance(dest, fld.Field) or dest is None:
            results = np.zeros(len(fkey_index_spans) - 1, dtype=result_dtype)
        else:
            results = dest

        # execute the predicate (note that not every predicate requires a reader)
        predicate(fkey_index_spans, src, results)

        if isinstance(dest, fld.Field):
            dest.data.write(results)

        return dest if dest is not None else results


    def aggregate_count(self, fkey_indices=None, fkey_index_spans=None,
                        src=None, dest=None):
        return per._aggregate_impl(self.apply_spans_count, fkey_indices, fkey_index_spans,
                                   src, dest, np.uint32)


    def aggregate_first(self, fkey_indices=None, fkey_index_spans=None,
                        src=None, dest=None):
        return self.aggregate_custom(self.apply_spans_first, fkey_indices, fkey_index_spans,
                                     src, dest)


    def aggregate_last(self, fkey_indices=None, fkey_index_spans=None,
                       src=None, dest=None):
        return self.aggregate_custom(self.apply_spans_last, fkey_indices, fkey_index_spans,
                                     src, dest)


    def aggregate_min(self, fkey_indices=None, fkey_index_spans=None,
                      src=None, dest=None):
        return self.aggregate_custom(self.apply_spans_min, fkey_indices, fkey_index_spans,
                                     src, dest)


    def aggregate_max(self, fkey_indices=None, fkey_index_spans=None,
                      src=None, dest=None):
        return self.aggregate_custom(self.apply_spans_max, fkey_indices, fkey_index_spans,
                                     src, dest)


    def aggregate_custom(self,
                         predicate, fkey_indices=None, fkey_index_spans=None,
                         src=None, dest=None):
        if src is None:
            raise ValueError("'reader' must not be None")
        if not isinstance(src, (rw.Reader, np.ndarray)):
            raise ValueError(f"'reader' must be a Reader or an ndarray but is {type(src)}")
        required_dtype = src.dtype
        return per._aggregate_impl(predicate, fkey_indices, fkey_index_spans,
                                   src, dest, required_dtype)


    def join(self,
             destination_pkey, fkey_indices, values_to_join,
             writer=None, fkey_index_spans=None):
        if fkey_indices is not None:
            if not isinstance(fkey_indices, (rw.Reader, np.ndarray)):
                raise ValueError(f"'fkey_indices' must be a type of Reader or an ndarray")
        if values_to_join is not None:
            if not isinstance(values_to_join, (rw.Reader, np.ndarray)):
                raise ValueError(f"'values_to_join' must be a type of Reader but is {type(values_to_join)}")
            if isinstance(values_to_join, rw.IndexedStringReader):
                raise ValueError(f"Joins on indexed string fields are not supported")

        if isinstance(values_to_join, rw.Reader):
            raw_values_to_join = values_to_join[:]
        else:
            raw_values_to_join = values_to_join

        # generate spans for the sorted key indices if not provided
        if fkey_index_spans is None:
            fkey_index_spans = self.get_spans(field=fkey_indices)

        # select the foreign keys from the start of each span to get an ordered list
        # of unique id indices in the destination space that the results of the predicate
        # execution are mapped to
        unique_fkey_indices = fkey_indices[:][fkey_index_spans[:-1]]

        # generate a filter to remove invalid foreign key indices (where values in the
        # foreign key don't map to any values in the destination space
        invalid_filter = unique_fkey_indices < operations.INVALID_INDEX
        safe_unique_fkey_indices = unique_fkey_indices[invalid_filter]

        # the predicate results are in the same space as the unique_fkey_indices, which
        # means they may still contain invalid indices, so filter those now
        safe_values_to_join = raw_values_to_join[invalid_filter]

        # now get the memory that the results will be mapped to
        destination_space_values = writer.chunk_factory(len(destination_pkey))

        # finally, map the results from the source space to the destination space
        destination_space_values[safe_unique_fkey_indices] = safe_values_to_join

        if writer is not None:
            writer.write(destination_space_values)
        else:
            return destination_space_values


    def predicate_and_join(self,
                           predicate, destination_pkey, fkey_indices,
                           reader=None, writer=None, fkey_index_spans=None):
        if reader is not None:
            if not isinstance(reader, rw.Reader):
                raise ValueError(f"'reader' must be a type of Reader but is {type(reader)}")
            if isinstance(reader, rw.IndexedStringReader):
                raise ValueError(f"Joins on indexed string fields are not supported")

        # generate spans for the sorted key indices if not provided
        if fkey_index_spans is None:
            fkey_index_spans = self.get_spans(field=fkey_indices)

        # select the foreign keys from the start of each span to get an ordered list
        # of unique id indices in the destination space that the results of the predicate
        # execution are mapped to
        unique_fkey_indices = fkey_indices[:][fkey_index_spans[:-1]]

        # generate a filter to remove invalid foreign key indices (where values in the
        # foreign key don't map to any values in the destination space
        invalid_filter = unique_fkey_indices < operations.INVALID_INDEX
        safe_unique_fkey_indices = unique_fkey_indices[invalid_filter]

        # execute the predicate (note that not every predicate requires a reader)
        if reader is not None:
            dtype = reader.dtype()
        else:
            dtype = np.uint32
        results = np.zeros(len(fkey_index_spans)-1, dtype=dtype)
        predicate(fkey_index_spans, reader, results)

        # the predicate results are in the same space as the unique_fkey_indices, which
        # means they may still contain invalid indices, so filter those now
        safe_results = results[invalid_filter]

        # now get the memory that the results will be mapped to
        destination_space_values = writer.chunk_factory(len(destination_pkey))
        # finally, map the results from the source space to the destination space
        destination_space_values[safe_unique_fkey_indices] = safe_results

        writer.write(destination_space_values)

    def get(self, field):
        if 'fieldtype' not in field.attrs.keys():
            raise ValueError(f"'{field}' is not a well-formed field")

        fieldtype_map = {
            'indexedstring': fld.IndexedStringField,
            'fixedstring': fld.FixedStringField,
            'categorical': fld.CategoricalField,
            'boolean': fld.NumericField,
            'numeric': fld.NumericField,
            'datetime': fld.TimestampField,
            'date': fld.TimestampField,
            'timestamp': fld.TimestampField
        }

        fieldtype = field.attrs['fieldtype'].split(',')[0]
        return fieldtype_map[fieldtype](self, field)


    def create_like(self, field, dest_group, dest_name, timestamp=None, chunksize=None):
        if 'fieldtype' not in field.attrs.keys():
            raise ValueError("{} is not a well-formed field".format(field))

        f = self.get(field)
        f.create_like(dest_group, dest_name)


    def create_indexed_string(self, group, name, timestamp=None, chunksize=None):
        fld.indexed_string_field_constructor(self, group, name, timestamp, chunksize)
        return fld.IndexedStringField(self, group[name], write_enabled=True)


    def create_fixed_string(self, group, name, length, timestamp=None, chunksize=None):
        fld.fixed_string_field_constructor(self, group, name, length, timestamp, chunksize)
        return fld.FixedStringField(self, group[name], write_enabled=True)


    def create_categorical(self, group, name, nformat, key,
                           timestamp=None, chunksize=None):
        fld.categorical_field_constructor(self, group, name, nformat, key,
                                          timestamp, chunksize)
        return fld.CategoricalField(self, group[name], write_enabled=True)


    def create_numeric(self, group, name, nformat, timestamp=None, chunksize=None):
        fld.numeric_field_constructor(self, group, name, nformat, timestamp, chunksize)
        return fld.NumericField(self, group[name], write_enabled=True)


    def create_timestamp(self, group, name, timestamp=None, chunksize=None):
        fld.timestamp_field_constructor(self, group, name, timestamp, chunksize)
        return fld.TimestampField(self, group[name], write_enabled=True)


    def get_or_create_group(self, group, name):
        if name in group:
            return group[name]
        return group.create_group(name)


    def chunks(self, length, chunksize=None):
        if chunksize is None:
            chunksize = self.chunksize
        cur = 0
        while cur < length:
            next = min(length, cur + chunksize)
            yield cur, next
            cur = next


    def process(self, inputs, outputs, predicate):

        # TODO: modifying the dictionaries in place is not great
        input_readers = dict()
        for k, v in inputs.items():
            if isinstance(v, rw.Reader):
                input_readers[k] = v
            else:
                input_readers[k] = self.get_reader(v)
        output_writers = dict()
        output_arrays = dict()
        for k, v in outputs.items():
            if isinstance(v, rw.Writer):
                output_writers[k] = v
            else:
                raise ValueError("'outputs': all values must be 'Writers'")

        reader = next(iter(input_readers.values()))
        input_length = len(reader)
        writer = next(iter(output_writers.values()))
        chunksize = writer.chunksize
        required_chunksize = min(input_length, chunksize)
        for k, v in outputs.items():
            output_arrays[k] = output_writers[k].chunk_factory(required_chunksize)

        for c in self.chunks(input_length, chunksize):
            kwargs = dict()

            for k, v in inputs.items():
                kwargs[k] = v[c[0]:c[1]]
            for k, v in output_arrays.items():
                kwargs[k] = v[:c[1] - c[0]]
            predicate(**kwargs)

            # TODO: write back to the writer
            for k in output_arrays.keys():
                output_writers[k].write_part(kwargs[k])
        for k, v in output_writers.items():
            output_writers[k].flush()


    def get_index(self, target, foreign_key, destination=None):
        print('  building patient_id index')
        t0 = time.time()
        target_lookup = dict()
        for i, v in enumerate(target[:]):
            target_lookup[v] = i
        print(f'  target lookup built in {time.time() - t0}s')

        print('  perform initial index')
        t0 = time.time()
        foreign_key_elems = foreign_key[:]
        # foreign_key_index = np.asarray([target_lookup.get(i, -1) for i in foreign_key_elems],
        #                                    dtype=np.int64)
        foreign_key_index = np.zeros(len(foreign_key_elems), dtype=np.int64)

        current_invalid = np.int64(operations.INVALID_INDEX)
        for i_k, k in enumerate(foreign_key_elems):
            index = target_lookup.get(k, current_invalid)
            if index >= operations.INVALID_INDEX:
                current_invalid += 1
                target_lookup[k] = index
            foreign_key_index[i_k] = index
        print(f'  initial index performed in {time.time() - t0}s')

        if destination:
            destination.write(foreign_key_index)
        else:
            return foreign_key_index


    def get_trash_group(self, group):

        group_names = group.name[1:].split('/')

        while True:
            id = str(uuid.uuid4())
            try:
                result = group.create_group(f"/trash/{'/'.join(group_names[:-1])}/{id}")
                return result
            except KeyError:
                pass


    def temp_filename(self):
        uid = str(uuid.uuid4())
        while os.path.exists(uid + '.hdf5'):
            uid = str(uuid.uuid4())
        return uid + '.hdf5'


    def map_left(self, left_on, right_on, map_mode='both'):
        valid_map_modes = ('map_to', 'map_from', 'both')

        if map_mode not in valid_map_modes:
            raise ValueError("'map_mode' must be one of {}".format(valid_map_modes))

        l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
        l_index = np.arange(len(l_key_raw), dtype=np.int64)
        l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})

        r_key_raw = val.raw_array_from_parameter(self, 'right_on', right_on)
        r_index = np.arange(len(r_key_raw), dtype=np.int64)
        r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})

        df = pd.merge(left=l_df, right=r_df, left_on='l_k', right_on='r_k', how='left')
        l_to_r_map = df['l_index'].to_numpy()
        l_to_r_filt = np.logical_not(df['l_index'].isnull())
        r_to_l_map = df['r_index'].to_numpy(dtype=np.int64)
        r_to_l_filt = np.logical_not(df['r_index'].isnull())
        return l_to_r_map, l_to_r_filt, r_to_l_map, r_to_l_filt

    def map_right(self, left_on, right_on, map_mode='both'):
        l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
        l_index = np.arange(len(l_key_raw), dtype=np.int64)
        l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})

        r_key_raw = val.raw_array_from_parameter(self, 'right_on', right_on)
        r_index = np.arange(len(r_key_raw), dtype=np.int64)
        r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})

        df = pd.merge(left=r_df, right=l_df, left_on='r_k', right_on='l_k', how='left')
        l_to_r_map = df['r_index'].to_numpy()
        l_to_r_filt = np.logical_not(df['r_index'].isnull())
        r_to_l_map = df['l_index'].to_numpy(dtype=np.int64)
        r_to_l_filt = np.logical_not(df['l_index'].isnull())
        return l_to_r_map, l_to_r_filt, r_to_l_map, r_to_l_filt


    def merge(self, left_on, right_on, how,
              left_fields=None, right_fields=None,
              left_writers=None, right_writers=None):
        if how == 'left':
            self.merge_left(left_on, right_on, left_fields, right_fields,
                            left_writers, right_writers)
        elif how == 'right':
            self.merge_right(left_on, right_on, left_fields, right_fields,
                             left_writers, right_writers)
        elif how == 'inner':
            self.merge_inner(left_on, right_on, left_fields, right_fields,
                             left_writers, right_writers)
        elif how == 'outer':
            self.merge_outer(left_on, right_on, left_fields, right_fields,
                             left_writers, right_writers)


    def merge_left(self, left_on, right_on,
                   left_fields=tuple(), right_fields=tuple(),
                   left_writers=None, right_writers=None):
        l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
        l_index = np.arange(len(l_key_raw), dtype=np.int64)
        l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})

        r_key_raw = val.raw_array_from_parameter(self, 'right_on', right_on)
        r_index = np.arange(len(r_key_raw), dtype=np.int64)
        r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})

        df = pd.merge(left=l_df, right=r_df, left_on='l_k', right_on='r_k', how='left')
        l_to_r_map = df['l_index'].to_numpy()
        l_to_r_filt = np.logical_not(df['l_index'].isnull()).to_numpy()
        r_to_l_map = df['r_index'].to_numpy(dtype=np.int64)
        r_to_l_filt = np.logical_not(df['r_index'].isnull()).to_numpy()
        # print(df)

        # print("l_to_r_map:", l_to_r_map, l_to_r_map.dtype)
        # print("r_to_l_map:", r_to_l_map, r_to_l_map.dtype)

        # print(r_to_l_map > -1)
        left_results = list()
        for ilf, lf in enumerate(left_fields):
            lf_raw = val.raw_array_from_parameter(self, 'left_fields[{}]'.format(ilf), lf)
            joined_field = per._safe_map(lf_raw, l_to_r_map, l_to_r_filt)
            print(joined_field)
            if left_writers == None:
                left_results.append(joined_field)
            else:
                left_writers[ilf].write(joined_field)

        right_results = list()
        for irf, rf in enumerate(right_fields):
            rf_raw = val.raw_array_from_parameter(self, 'right_fields[{}]'.format(irf), rf)
            joined_field = per._safe_map(rf_raw, r_to_l_map, r_to_l_filt)
            print(joined_field)
            if right_writers == None:
                right_results.append(joined_field)
            else:
                right_writers[irf].write(joined_field)

        if left_writers == None:
            return left_results, right_results


    def merge_right(self, left_on, right_on,
                    left_fields=None, right_fields=None,
                    left_writers=None, right_writers=None):
        l_key_raw = val.raw_array_from_parameter(self, 'left_on', left_on)
        l_index = np.arange(len(l_key_raw), dtype=np.int64)
        l_df = pd.DataFrame({'l_k': l_key_raw, 'l_index': l_index})

        r_key_raw = val.raw_array_from_parameter(self, 'right_on', right_on)
        r_index = np.arange(len(r_key_raw), dtype=np.int64)
        r_df = pd.DataFrame({'r_k': r_key_raw, 'r_index': r_index})

        df = pd.merge(left=r_df, right=l_df, left_on='r_k', right_on='l_k', how='left')
        l_to_r_map = df['r_index'].to_numpy()
        l_to_r_filt = np.logical_not(df['r_index'].isnull())
        r_to_l_map = df['l_index'].to_numpy(dtype=np.int64)
        r_to_l_filt = np.logical_not(df['l_index'].isnull())
        print(df)

        print("l_to_r_map:", l_to_r_map, l_to_r_map.dtype)
        print("r_to_l_map:", r_to_l_map, r_to_l_map.dtype)

        print(r_to_l_map > -1)
        left_results = list()
        for ilf, lf in enumerate(left_fields):
            joined_field = per._safe_map(lf, r_to_l_map)
            print(joined_field)
            if left_writers == None:
                left_results.append(joined_field)
            else:
                left_writers[ilf].write(joined_field)

        right_results = list()
        for irf, rf in enumerate(right_fields):
            joined_field = per._safe_map(rf, l_to_r_map)
            print(joined_field)
            if right_writers == None:
                right_results.append(joined_field)
            else:
                right_writers[irf].write(joined_field)


    def merge_inner(self, left_on, left_fields, right_on, right_fields):
        raise NotImplementedError()


    def merge_outer(self, left_on, left_fields, right_on, right_fields):
        raise NotImplementedError()


    def _map_fields(self, field_map, field_sources, field_sinks):
        rtn_sinks = None
        if field_sinks is None:
            left_sinks = list()
            for src in field_sources:
                src_ = val.array_from_parameter(self, 'left_field_sources', src)
                snk_ = ops.map_valid(src_, field_map)
                left_sinks.append(snk_)
            rtn_sinks = left_sinks
        elif val.is_field_parameter(field_sinks[0]):
            # groups or fields
            for src, snk in zip(field_sources, field_sinks):
                with utils.Timer("reading"):
                    src_ = val.array_from_parameter(self, 'left_field_sources', src)
                with utils.Timer("mapping"):
                    snk_ = ops.map_valid(src_, field_map)
                with utils.Timer("getting_sink"):
                    snk = val.field_from_parameter(self, 'left_field_sinks', snk)
                with utils.Timer("writing"):
                    snk.data.write(snk_)
                # src_ = val.array_from_parameter(self, 'left_field_sources', src)
                # snk_ = ops.map_valid(src_, field_map)
                # snk = val.field_from_parameter(self, 'left_field_sinks', snk)
                # snk.data.write(snk_)
        else:
            # raw arrays
            for src, snk in zip(field_sources, field_sinks):
                src_ = val.array_from_parameter(self, 'left_field_sources', src)
                snk_ = val.array_from_parameter(self, 'left_field_sinks', snk)
                ops.map_valid(src_, field_map, snk_)
        return rtn_sinks


    def ordered_merge(self, left_on, right_on, how,
                      left_field_sources=tuple(), left_field_sinks=None,
                      right_field_sources=tuple(), right_field_sinks=None,
                      left_unique=False, right_unique=False):
        if how == 'left':
            return self.ordered_left_merge(left_on, right_on,
                                           left_field_sources, left_field_sinks,
                                           left_unique, right_unique)
        elif how == 'right':
            return self.ordered_left_merge(right_on, left_on,
                                           right_field_sources, right_field_sinks,
                                           right_unique, left_unique)
        elif how == 'inner':
            return self.ordered_inner_merge(left_on, right_on,
                                            left_field_sources, left_field_sinks,
                                            right_field_sources, right_field_sinks,
                                            left_unique, right_unique)
        else:
            raise ValueError("'how' must be one of 'left', 'right' or 'inner'")


    def ordered_left_merge(self, left_on, right_on, left_to_right_map,
                           left_field_sources=tuple(), left_field_sinks=None,
                           left_unique=False, right_unique=False):

        if left_field_sinks is not None:
            if len(left_field_sources) != len(left_field_sinks):
                msg = ("{} and {} should be of the same length but are length {} and {} "
                       "respectively")
                raise ValueError(msg.format(len(left_field_sources), len(left_field_sinks)))
        val.all_same_basic_type('left_field_sources', left_field_sources)
        if left_field_sinks and len(left_field_sinks) > 0:
            val.all_same_basic_type('left_field_sinks', left_field_sinks)

        streamable = val.is_field_parameter(left_on) and \
                     val.is_field_parameter(right_on) and \
                     val.is_field_parameter(left_field_sources[0]) and \
                     left_field_sinks is not None and \
                     val.is_field_parameter(left_field_sinks[0])

        with utils.Timer("generating maps"):
            result = None
            has_unmapped = None
            if left_unique == False:
                if right_unique == False:
                    raise NotImplementedError()
                else:
                    raise NotImplementedError()
            else:
                if right_unique == False:
                    if streamable:
                        has_unmapped =\
                            ops.ordered_map_to_right_left_unique_streamed(left_on, right_on,
                                                                          left_to_right_map)
                        result = left_to_right_map.data[:]
                    else:
                        result = np.zeros(len(right_on), dtype=np.int64)
                        has_unmapped =\
                            ops.ordered_map_to_right_left_unique(left_on, right_on, result)

                else:
                    result = np.zeros(len(right_on), dtype=np.int64)
                    has_unmapped = ops.ordered_map_to_right_both_unique(left_on, right_on, result)

        with utils.Timer("mapping fields", new_line=True):
            rtn_left_sinks = self._map_fields(result, left_field_sources, left_field_sinks)
        return rtn_left_sinks


    def ordered_right_merge(self, left_on, right_on,
                            right_field_sources=tuple(), right_field_sinks=None,
                            left_unique=False, right_unique=False):
        return self.ordered_left_merge(right_on, left_on,
                                       right_field_sources, right_field_sinks,
                                       right_unique, left_unique)


    def ordered_inner_merge(self, left_on, right_on,
                            left_field_sources=tuple(), left_field_sinks=None,
                            right_field_sources=tuple(), right_field_sinks=None,
                            left_unique=False, right_unique=False):

        if left_field_sinks is not None:
            if len(left_field_sources) != len(left_field_sinks):
                msg = ("{} and {} should be of the same length but are length {} and {} "
                       "respectively")
                raise ValueError(msg.format(len(left_field_sources), len(left_field_sinks)))
        val.all_same_basic_type('left_field_sources', left_field_sources)
        if left_field_sinks and len(left_field_sinks) > 0:
            val.all_same_basic_type('left_field_sinks', left_field_sinks)

        if right_field_sinks is not None:
            if len(right_field_sources) != len(right_field_sinks):
                msg = ("{} and {} should be of the same length but are length {} and {} "
                       "respectively")
                raise ValueError(msg.format(len(right_field_sources), len(right_field_sinks)))
        val.all_same_basic_type('right_field_sources', right_field_sources)
        if right_field_sinks and len(right_field_sinks) > 0:
            val.all_same_basic_type('right_field_sinks', right_field_sinks)

        result = None
        if left_unique is False:
            if right_unique is False:
                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
                left_to_inner = np.zeros(inner_length, dtype=np.int64)
                right_to_inner = np.zeros(inner_length, dtype=np.int64)
                ops.ordered_inner_map(left_on, right_on, left_to_inner, right_to_inner)
            else:
                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
                left_to_inner = np.zeros(inner_length, dtype=np.int64)
                right_to_inner = np.zeros(inner_length, dtype=np.int64)
                ops.ordered_inner_map_left_unique(right_on, left_on, right_to_inner, left_to_inner)
        else:
            if right_unique is False:
                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
                left_to_inner = np.zeros(inner_length, dtype=np.int64)
                right_to_inner = np.zeros(inner_length, dtype=np.int64)
                ops.ordered_inner_map_left_unique(left_on, right_on, left_to_inner, right_to_inner)
            else:
                inner_length = ops.ordered_inner_map_result_size(left_on, right_on)
                left_to_inner = np.zeros(inner_length, dtype=np.int64)
                right_to_inner = np.zeros(inner_length, dtype=np.int64)
                ops.ordered_inner_map_both_unique(left_on, right_on, left_to_inner, right_to_inner)

        rtn_left_sinks = self._map_fields(left_to_inner, left_field_sources, left_field_sinks)
        rtn_right_sinks = self._map_fields(right_to_inner, right_field_sources, right_field_sinks)

        if rtn_left_sinks:
            if rtn_right_sinks:
                return rtn_left_sinks, rtn_right_sinks
            else:
                return rtn_left_sinks
        else:
            return rtn_right_sinks



from collections import defaultdict
from datetime import datetime, timezone
import numpy as np
import h5py

from hystore.core import persistence
from hystore.core.utils import build_histogram
from hystore.core import exporter
filename = '/home/ben/covid/ds_20200830_full.hdf5'
filenametemp = '/home/ben/covid/ds_temp.hdf5'
ds = persistence.DataStore()
with h5py.File(filename, 'r') as src:
    with h5py.File(filenametemp, 'w') as tmp:

        s_p = src['patients']
        print(s_p.keys())

        for k in ('lsoa11cd', 'lsoa11nm', 'msoa11cd', 'msoa11nm', 'ladcd', 'outward_postcode_region'):
            d = defaultdict(int)
            f = ds.get_reader(s_p[k])[:]
            for v in f:
                d[v] += 1
            print(k, sorted(d.items(), key=lambda x: x[1], reverse=True)[:10])




# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

class NumpyBuffer2:
    def __init__(self, dtype, block_pow=8):
        self.dtype_ = dtype
        self.list_ = list()

    def append(self, value):
        self.list_.append(value)

    def finalise(self):
        result = np.asarray(self.list_, dtype=self.dtype_)
        del self.list_
        return result


class NumpyBuffer:
    def __init__(self, dtype, block_pow=8):
        self.block_shift_ = block_pow
        self.block_size_ = 1 << block_pow
        self.block_mask_ = self.block_size_ - 1
        self.blocks_ = list()
        self.dtype_ = dtype
        self.current_block_ = None
        self.current_ = 0

    def append(self, value):
        if self.current_ == len(self.blocks_) * self.block_size_:
            self.blocks_.append(np.zeros(self.block_size_, dtype=self.dtype_))
            self.current_block_ = self.blocks_[-1]
        self.current_block_[self.current_ & self.block_mask_] = value
        self.current_ += 1


    def finalise(self):
        if self.current_ == 0:
            return None
        final = np.zeros(self.current_, dtype=self.dtype_)
        start = 0
        for b in range(len(self.blocks_) - 1):
            cur_block = self.blocks_[b]
            final[start:start + self.block_size_] = cur_block
            start += self.block_size_
            del cur_block
        current = self.current_ & self.block_mask_
        last_block = self.blocks_[-1]
        final[start:start + current] = last_block[0:current]
        del last_block

        self.blocks_ = list()
        self.current_ = 0

        return final


class ListBuffer:
    def __init__(self, block_pow=8):
        self.block_shift_ = block_pow
        self.block_size_ = 1 << block_pow
        self.block_mask_ = self.block_size_ - 1
        self.blocks_ = list()
        self.current_ = 0

    def append(self, value):
        if self.current_ == len(self.blocks_) * self.block_size_:
            self.blocks_.append([None] * self.block_size_)
        self.blocks_[self.current_ >> self.block_shift_][self.current_ & self.block_mask_] = value
        self.current_ += 1


    def finalise(self):
        if self.current_ == 0:
            return None
        final = [None] * self.current_
        start = 0
        for b in range(len(self.blocks_) - 1):
            cur_block = self.blocks_[b]
            final[start:start + self.block_size_] = cur_block
            start += self.block_size_
            del cur_block
        current = self.current_ & self.block_mask_
        last_block = self.blocks_[-1]
        final[start:start + current] = last_block[0:current]
        del last_block

        self.blocks_ = list()
        self.current_ = 0

        return final

# x = np.zeros(32)
# y = np.asarray([x for x in range(16)])
# start = 0
# inc = 16
# x[start:start+inc] = y
# print(y)


import numpy as np

from hystore.core import utils


def method_paper_model(ds, symptoms_reader_dict, prediction):

    intercept = -1.19015973
    weights = {'persistent_cough': 0.23186655,
               'fatigue': 0.56532346,
               'delirium': -0.12935112,
               'shortness_of_breath': 0.58273967,
               'fever': 0.16580974,
               'diarrhoea': 0.10236126,
               'abdominal_pain': -0.11204163,
               'chest_pain': -0.12318634,
               'hoarse_voice': -0.17818597,
               'skipped_meals': 0.25902482,
               'loss_of_smell': 1.82895239}

    with utils.Timer("predicting covid by assessment", new_line=True):
        cumulative = np.zeros(len(symptoms_reader_dict['persistent_cough']), dtype='float32')
        for s in symptoms_reader_dict:
            cumulative += symptoms_reader_dict[s][:] * weights[s]
        cumulative += intercept
        prediction.write(cumulative)



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import argparse
import os
import datetime

from hystore.covidspecific import data_schemas, parsing_schemas, pipeline
from hystore.core import dataset, utils

equivalence_map = {
    'na': ('', 'na'),
    '': ('', 'na')
}

class Logger(object):
    def __init__(self):
        self.report = ''

    def log(self, text, print_to_console=False):
        if isinstance(text, list):
            for item in text:
                self.report += str(item)
                self.report += '\n'
                if print_to_console:
                    print(item)
        else:
            self.report += text
            self.report += '\n'
            if print_to_console:
                print(text)

    def write(self, destination, append=False):
        if os.path.isfile(destination) and append is False:
            mode = 'w'
        else:
            mode = 'a'
        with open(destination, mode) as f:
            f.write(self.report)




def compare_row(expected_index, expected_row, actual_index, actual_row, keys):
    all_true = True
    mismatches = None
    for k in keys:
        expected_value = expected_row.value_from_fieldname(expected_index, k)
        actual_value = actual_row.value_from_fieldname(actual_index, k)
        if expected_value in equivalence_map:
            if actual_value not in equivalence_map[expected_value]:
                all_true = False
                if mismatches is None:
                    mismatches = list()
                mismatches.append(k)
            # all_true = all_true and actual_value in equivalence_map[expected_value]
        else:
            if expected_value != actual_value:
                all_true = False
                if mismatches is None:
                    mismatches = list()
                mismatches.append(k)
            # all_true = all_true and expected_value == actual_value
    return all_true, mismatches


def compare(expected, actual, verbose):
    logger = Logger()
    logger.log(f'\nComparing expected {expected} and current {actual}', print_to_console=True)

    with open(expected) as f:
        expected_ds = dataset.Dataset(f)
    expected_field_names = expected_ds.names_
    with open(actual) as f:
        actual_ds = dataset.Dataset(f)
    actual_field_names = actual_ds.names_

    # compare field names
    if expected_field_names == actual_field_names:
        logger.log('field names are identical')
    else:
        logger.log(['WARNING: fieldnames only in expected:', set(expected_field_names).difference(set(actual_field_names))], print_to_console = True)
        logger.log(['WARNING: fieldnames only in actual:', set(actual_field_names).difference(set(expected_field_names))], print_to_console = True)

    common_fields = set(expected_field_names).intersection(set(actual_field_names))
    if verbose:
        logger.log(['common fields:', common_fields,'\n'], print_to_console = True)

    # compare data content
    if expected_ds.row_count() == actual_ds.row_count():
        logger.log(f'both files have {expected_ds.row_count()} rows', print_to_console=True)
    else:
        logger.log(f'WARNING: expected file has {expected_ds.row_count()} rows, actual file has {actual_ds.row_count()} rows', print_to_console=True)


    rows_that_differ = []
    for i in range(expected_ds.row_count()):
        matched, mismatches = compare_row(i, expected_ds, i, actual_ds, common_fields)
        if not matched:
            rows_that_differ.append((i, mismatches))


    if len(rows_that_differ) > 0:
        logger.log(f'WARNING: a total of {len(rows_that_differ)} rows do not match', print_to_console=True)
        if verbose:
            for row in rows_that_differ:
                logger.log('expected value | actual_value')
                for field in row[1]:# common_fields:
                    logger.log(f'{expected_ds.value_from_fieldname(row[0], field)} | {actual_ds.value_from_fieldname(row[0], field)}')
    else:
        logger.log('row contents are equal \n \n', print_to_console=True)
    return logger


def validate(expected_patients_file_name, actual_patients_file_name,
             expected_assessments_file_name, actual_assessments_file_name,
             verbose=False, report_dest=None):
    log = compare(expected_patients_file_name, actual_patients_file_name, verbose)
    if report_dest is not None:
        log.write(report_dest)
    log = compare(expected_assessments_file_name, actual_assessments_file_name, verbose)
    if report_dest is not None:
        log.write(report_dest, append=True)


def validate_full():
    filename = '/home/ben/covid/assessments_20200413050002_clean.csv'

    with open(filename) as asmt:
        csvr = csv.DictReader(asmt)
        fieldnames = csvr.fieldnames
        print(fieldnames)

        for ir, r in enumerate(csvr):
            if r['tested_covid_positive'] != '':
                print(ir, r['id'], r['patient_id'], r['tested_covid_positive'])


def show_rows(filename, fields_to_show):
    with open(filename) as f:
        ds = dataset.Dataset(f)
    # ds.parse_file()
    ds.sort(('patient_id', 'updated_at'))
    for i_f, f in enumerate(ds.fields_):
        if ds.value_from_fieldname(i_f, 'patient_id') == 'e88bfabe5b16897866f91deb8a7a90f2':
            utils.print_diagnostic_row(f'{i_f}:', ds, ds.fields_, i_f, fields_to_show)


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('-pe', '--patients_output_expected',
                        help='the location of the patient file produced by a correct version (currently v.0.1.2) of the pipeline')
    parser.add_argument('-pi', '--patients_input',
                        help='the location of the patient input file to be processed')
    parser.add_argument('-ae', '--assessments_output_expected',
                        help='the location of the assessments file produced by a correct version (currently v.0.1.2) of the pipeline')
    parser.add_argument('-ai', '--assessments_input',
                        help='the location of the assessments input file to be processed')
    parser.add_argument('-r', '--report_destination',
                        help='the location for the report .txt to be saved',
                        default=None)
    parser.add_argument('-ps', '--parsing_schema', default=1,
                        help='the schema number to use for parsing and cleaning data')
    parser.add_argument('-y', '--year', default=datetime.datetime.now().year, type=int)
    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help='increase verbosity')
    args = parser.parse_args()

    data_schema_version = 1
    data_schema = data_schemas.DataSchema(data_schema_version)
    parsing_schema_version = args.parsing_schema
    parsing_schema = parsing_schemas.ParsingSchema(parsing_schema_version)
    # run the current pipeline
    pipeline_output = pipeline.pipeline(patient_filename=args.patients_input,
                                        assessment_filename=args.assessments_input,
                                        data_schema=data_schema, parsing_schema=parsing_schema,
                                        year=args.year)
    pipeline.save_csv(pipeline_output,
                      patient_data_out='patients_output_test.csv',
                      assessment_data_out='assessments_output_test.csv',
                      data_schema=data_schema)

    # compare to expected pipeline output
    validate(expected_patients_file_name=args.patients_output_expected,
             actual_patients_file_name='patients_output_test.csv',
             expected_assessments_file_name=args.assessments_output_expected,
             actual_assessments_file_name='assessments_output_test.csv',
             report_dest=args.report_destination,
             verbose=args.verbose)


from collections import defaultdict

import numpy as np


from hystore.core import dataset, utils

# keys = ('health_worker_with_contact', 'asmt_prediction', 'old_test_result', 'new_test_result')

meanings = {
    ('0', '3'): 1,
    ('0', '4'): 2,
    ('1', '3'): 1,
    ('1', '4'): 2,
    ('2', '0'): 1,
    ('2', '1'): 1,
    ('2', '2'): 1,
    ('2', '3'): 1,
    ('2', '4'): 2,
    ('3', '0'): 2,
    ('3', '1'): 2,
    ('3', '2'): 2,
    ('3', '3'): 2,
    ('3', '4'): 2
}

#filename = '/home/ben/OneDrive/supplement_paper/supplements_patients_20200731_1.csv'
filename = '/home/ben/covid/supplements_patients.csv'
with open(filename) as f:
    # ds = dataset.Dataset(f, keys=keys, show_progress_every=1000000)
    ds = dataset.Dataset(f, show_progress_every=250000)
    print(sorted(utils.build_histogram(ds.field_by_name('asmt_prediction'))))
    print(sorted(utils.build_histogram(ds.field_by_name('new_test_result'))))
    print(sorted(utils.build_histogram(ds.field_by_name('old_test_result'))))

    print(sorted(utils.build_histogram(ds.field_by_name('health_worker_with_contact'))))

    categories = defaultdict(int)

    otr = ds.field_by_name('old_test_result')
    ntr = ds.field_by_name('new_test_result')


    for i_r in range(len(otr)):
        categories[(otr[i_r], ntr[i_r])] += 1

    results = sorted(list(categories.items()))
    total_negative = 0
    total_positive = 0
    for r in results:
        print(r)
        if r[0] in meanings:
            if meanings[r[0]] == 1:
                total_negative += r[1]
            else:
                total_positive += r[1]

    print(total_negative)
    print(total_positive)

    asmt_keys = [n for n in ds.names_ if "asmt" in n]
    asmt_keys = [n for n in asmt_keys if n != 'asmt_prediction_score']
    for k in asmt_keys:
        f = ds.field_by_name(k)
        print(k, utils.build_histogram(sorted(f)))



import numpy as np

import h5py

from hystore.core import persistence as per
from hystore.core import fields as fld
from hystore.core import readerwriter as rw


def _writer_from_writer_or_group(writer_getter, param_name, writer):
    if isinstance(writer, h5py.Group):
        return writer_getter.get_existing_writer(writer)
    elif isinstance(writer, rw.Writer):
        return writer
    else:
        msg = "'{}' must be one of (h5py.Group, Writer) but is {}"
        raise ValueError(msg.format(param_name, type(writer)))


def _check_is_appropriate_writer_if_set(reader_getter, param_name, reader, writer):
    # TODO: this method needs reworking; readers should know whether writers are compatible with
    # them
    msg = "{} must be of type {} or None but is {}"
    # if writer is not None:
    #     if isinstance(reader, np.ndarray):
    #         raise ValueError("'if 'reader' is a numpy.ndarray, 'writer' must be None")

    if isinstance(reader, h5py.Group):
        reader = reader_getter.get_reader(reader)


    if isinstance(reader, rw.IndexedStringReader):
        if not isinstance(writer, rw.IndexedStringWriter):
            raise ValueError(msg.format(param_name, rw.IndexedStringReader, writer))
    elif isinstance(reader, rw.FixedStringReader):
        if not isinstance(writer, rw.FixedStringWriter):
            raise ValueError(msg.format(param_name, rw.FixedStringReader, writer))
    elif isinstance(reader, rw.NumericReader):
        if not isinstance(writer, rw.NumericWriter):
            raise ValueError(msg.format(param_name, rw.NumericReader, writer))
    elif isinstance(reader, rw.CategoricalReader):
        if not isinstance(writer, rw.CategoricalWriter):
            raise ValueError(msg.format(param_name, rw.CategoricalReader, writer))
    elif isinstance(reader, rw.TimestampReader):
        if not isinstance(writer, rw.TimestampWriter):
            raise ValueError(msg.format(param_name, rw.TimestampReader, writer))


def _check_all_readers_valid_and_same_type(readers):
    if not isinstance(readers, (tuple, list)):
        raise ValueError("'readers' collection must be a tuple or list")

    if isinstance(readers[0], h5py.Group):
        expected_type = h5py.Group
    elif isinstance(readers[0], rw.Reader):
        expected_type = rw.Reader
    elif isinstance(readers[0], np.ndarray):
        expected_type = np.ndarray
    else:
        raise ValueError("'readers' collection must of the following types: "
                         "(h5py.Group, Reader, numpy.ndarray)")
    for r in readers[1:]:
        if not isinstance(r, expected_type):
            raise ValueError("'readers': all elements must be the same underlying type "
                             "(h5py.Group, Reader, numpy.ndarray")


def _check_is_reader_substitute(name, field):
    if not isinstance(field, (h5py.Group, rw.Reader, np.ndarray)):
        msg = "'{}' must be one of (h5py.Group, Reader, numpy.ndarray) but is '{}'"
        raise ValueError(msg.format(type(field)))


def _check_is_reader_or_ndarray(name, field):
    if not isinstance(field, (rw.Reader, np.ndarray)):
        raise ValueError(f"'name' must be either a Reader or an ndarray but is {type(field)}")


def _check_is_reader_or_ndarray_if_set(name, field):
    if not isinstance(field, (rw.Reader, np.ndarray)):
        raise ValueError("if set, 'name' must be either a Reader or an ndarray "
                         f"but is {type(field)}")


def _check_equal_length(name1, field1, name2, field2):
    if len(field1) != len(field2):
        msg = "'{}' must be the same length as '{}' (lengths {} and {} respectively)"
        raise ValueError(msg.format(name1, name2, len(field1), len(field2)))


def _reader_from_group_if_required(reader_source, name, reader):
    if isinstance(reader, h5py.Group):
        return reader_source.get_reader(reader)
    return reader


def raw_array_from_parameter(datastore, name, field):
    if isinstance(field, h5py.Group):
        return datastore.get_reader(field)[:]
    elif isinstance(field, rw.Reader):
        return field[:]
    elif isinstance(field, fld.Field):
        return field.data[:]
    elif isinstance(field, np.ndarray):
        return field
    else:
        error_str = "'{}' must be one of (Group, Reader, Field or ndarray, but is {}"
        raise ValueError(error_str.format(name, type(field)))


def array_from_parameter(session, name, field):
    if isinstance(field, h5py.Group):
        return session.get(field).data[:]
    elif isinstance(field, fld.Field):
        return field.data[:]
    elif isinstance(field, np.ndarray):
        return field
    else:
        error_str = "'{}' must be one of (Group, Field, or ndarray, but is {}"
        raise ValueError(error_str.format(name, type(field)))


def field_from_parameter(session, name, field):
    if isinstance(field, h5py.Group):
        return session.get(field)
    elif isinstance(field, fld.Field):
        return field
    else:
        error_str = "'{}' must be one of (Group, Field, or ndarray, but is {}"
        raise ValueError(error_str.format(name, type(field)))

def is_field_parameter(field):
    return isinstance(field, (fld.Field, h5py.Group))

def all_same_basic_type(name, fields):
    msg = "'{}' cannot be mixture of groups, fields and ndarrays".format(name)
    if isinstance(fields[0], h5py.Group):
        for f in fields[1:]:
            if not isinstance(f, h5py.Group):
                raise ValueError(msg)
    if isinstance(fields[0], fld.Field):
        for f in fields[1:]:
            if not isinstance(f, fld.Field):
                raise ValueError(msg)
    if isinstance(fields[0], np.ndarray):
        for f in fields[1:]:
            if not isinstance(f, np.ndarray):
                raise ValueError(msg)



import hystore
from hystore.core import readerwriter as rw


# field_writers = {
#     'idtype': lambda g, cs, n, ts: rw.FixedStringWriter(g, cs, n, 32, ts),
#     'datetimetype':
#         lambda g, cs, n, ts: rw.DateTimeImporter(g, cs, n, False, ts),
#     'optionaldatetimetype':
#         lambda g, cs, n, ts: rw.DateTimeImporter(g, cs, n, True, ts),
#     'datetype':
#         lambda g, cs, n, ts: rw.OptionalDateImporter(g, cs, n, False, ts),
#     'optionaldatetype':
#         lambda g, cs, n, ts: rw.OptionalDateImporter(g, cs, n, True, ts),
#     'versiontype': lambda g, cs, n, ts: rw.FixedStringWriter(g, cs, n, 10, ts),
#     'indexedstringtype': lambda g, cs, n, ts: rw.IndexedStringWriter(g, cs, n, ts),
#     'countrycodetype': lambda g, cs, n, ts: rw.FixedStringWriter(g, cs, n, 2, ts),
#     'unittype': lambda g, cs, n, ts: rw.FixedStringWriter(g, cs, n, 1, ts),
#     'categoricaltype':
#         lambda g, cs, n, stv, ts: rw.CategoricalWriter(g, cs, n, stv, ts),
#     'leakycategoricaltype':
#         lambda g, cs, n, stv, oor, ts: rw.LeakyCategoricalImporter(g, cs, n, stv,
#                                                                    oor, ts),
#     'float32type': lambda g, cs, n, ts: rw.NumericImporter(
#         g, cs, n, 'float32', pers.try_str_to_float, ts),
#     'uint16type': lambda g, cs, n, ts: rw.NumericImporter(
#         g, cs, n, 'uint16', pers.try_str_to_int, ts),
#     'yeartype': lambda g, cs, n, ts: rw.NumericImporter(
#         g, cs, n, 'uint32', pers.try_str_to_float_to_int, ts),
#     'geocodetype': lambda g, cs, n, ts: rw.FixedStringWriter(g, cs, n, 9, ts)
# }

# new_field_writers = {
#     'idtype': lambda s, g, n, ts=None, cs=None:
#     fields.FixedStringImporter(s, g, n, 32, ts, cs),
#     'datetimetype': lambda s, g, n, ts=None, cs=None:
#     fields.DateTimeImporter(s, g, n, False, True, ts, cs),
#     'optionaldatetimetype': lambda s, g, n, ts=None, cs=None:
#     fields.DateTimeImporter(s, g, n, True, True, ts, cs),
#     'datetype': lambda s, g, n, ts=None, cs=None:
#     fields.DateImporter(s, g, n, False, ts, cs),
#     'optionaldatetype': lambda s, g, n, ts=None, cs=None:
#     fields.DateImporter(s, g, n, True, ts, cs),
#     'versiontype': lambda s, g, n, ts=None, cs=None:
#     fields.FixedStringImporter(s, g, n, 10, ts, cs),
#     'indexedstringtype': lambda s, g, n, ts=None, cs=None:
#     fields.IndexedStringImporter(s, g, n, ts, cs),
#     'countrycodetype': lambda s, g, n, ts=None, cs=None:
#     fields.FixedStringImporter(s, g, n, 2, ts, cs),
#     'unittype': lambda s, g, n, ts=None, cs=None:
#     fields.FixedStringImporter(s, g, n, 1, ts, cs),
#     'categoricaltype': lambda s, g, n, vt, stv, ts=None, cs=None:
#     fields.CategoricalImporter(s, g, n, vt, stv, ts, cs),
#     'leakycategoricaltype': lambda s, g, n, vt, stv, oor, ts=None, cs=None:
#     fields.LeakyCategoricalImporter(s, g, n, vt, stv, oor, ts, cs),
#     'float32type': lambda s, g, n, ts=None, cs=None:
#     fields.NumericImporter(s, g, n, 'float32', pers.try_str_to_float, ts, cs),
#     'uint16type': lambda s, g, n, ts=None, cs=None:
#     fields.NumericImporter(s, g, n, 'uint16', pers.try_str_to_int, ts, cs),
#     'yeartype': lambda s, g, n, ts=None, cs=None:
#     fields.NumericImporter(s, g, n, 'uint32', pers.try_str_to_float_to_int, ts, cs),
#     'geocodetype': lambda s, g, n, ts=None, cs=None:
#     fields.FixedStringImporter(s, g, n, 9, ts, cs)
# }


new_field_importers = {
    'string': lambda:
        lambda ds, g, n, ts: rw.IndexedStringWriter(ds, g, n, ts),
    'fixed_string': lambda strlen:
        lambda ds, g, n, ts: rw.FixedStringWriter(ds, g, n, strlen, ts),
    'datetime': lambda optional:
        lambda ds, g, n, ts: rw.DateTimeImporter(ds, g, n, optional, ts),
    'date': lambda optional:
        lambda ds, g, n, ts: rw.OptionalDateImporter(ds, g, n, optional, ts),
    'numeric': lambda typestr, parser:
        lambda ds, g, n, ts: rw.NumericImporter(ds, g, n, typestr, parser, ts),
    'categorical': lambda stv, oor=None:
        lambda ds, g, n, ts: rw.CategoricalWriter(ds, g, n, stv, ts) if oor is None else
        rw.LeakyCategoricalImporter(ds, g, n, stv, oor, ts)
}


class FieldDesc:
    def __init__(self, field, importer, strings_to_values, values_to_strings, to_datatype,
                 out_of_range_label):
        self.field = field
        self.importer = importer
        self.to_datatype = to_datatype
        self.strings_to_values = strings_to_values
        self.values_to_strings = values_to_strings
        self.out_of_range_label = out_of_range_label

    def __str__(self):
        output = 'FieldDesc(field={}, strings_to_values={}, values_to_strings={})'
        return output.format(self.field, self.strings_to_values, self.values_to_strings)

    def __repr__(self):
        return self.__str__()


class FieldEntry:
    def __init__(self, field_desc, version_from, version_to=None):
        self.field_desc = field_desc
        self.version_from = version_from
        self.version_to = version_to

    def __str__(self):
        output = 'FieldEntry(field_desc={}, version_from={}, version_to={})'
        return output.format(self.field_desc, self.version_from, self.version_to)

    def __repr__(self):
        return self.__str__()



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import csv
import datetime
import time
from collections import defaultdict

import numpy as np

from hystore.core import dataset, filtered_field, regression
from hystore.covidspecific import parsing_schemas
from hystore.processing.age_from_year_of_birth import CalculateAgeFromYearOfBirth
from hystore.processing.assessment_merge import CalculateMergedFieldCount, MergeAssessmentRows
from hystore.processing.inconsistent_symptoms import CheckInconsistentSymptoms
from hystore.processing.inconsistent_testing import CheckTestingConsistency
from hystore.core import load_schema

from hystore.core.utils import count_flag_empty, count_flag_set, build_histogram, map_between_categories, \
    to_categorical, print_diagnostic_row, valid_range_fac_inc, datetime_to_seconds, concatenate_maybe_strs
from hystore.covidspecific.utils import iterate_over_patient_assessments, iterate_over_patient_assessments2


def copy_field(field):
    if isinstance(field, list):
        return copy.deepcopy(field)
    else:
        return field.copy()


def map_patient_ids(geoc, asmt, map_fn):
    g = 0
    a = 0
    while g < len(geoc) and a < len(asmt):
        gpid = geoc[g][0]
        apid = asmt[a][1]
        if gpid < apid:
            g += 1
        elif apid < gpid:
            a += 1
        else:
            map_fn(geoc, asmt, g, a)
            a += 1


#patient limits
MIN_AGE = 16
MAX_AGE = 90
MIN_HEIGHT = 110
MAX_HEIGHT = 220
MIN_WEIGHT = 40
MAX_WEIGHT = 200
MIN_BMI = 15
MAX_BMI = 55

MIN_TEMP = 35
MAX_TEMP = 42

# patient filter values
PFILTER_OTHER_TERRITORY = 0x1
# PFILTER_NO_ASSESSMENTS = 0x2
# PFILTER_ONE_ASSESSMENT = 0x4
FILTER_MISSING_AGE = 0x8
FILTER_BAD_AGE = 0x10
FILTER_MISSING_HEIGHT = 0x20
FILTER_BAD_HEIGHT = 0x40
FILTER_MISSING_WEIGHT = 0x80
FILTER_BAD_WEIGHT = 0x100
FILTER_MISSING_BMI = 0x200
FILTER_BAD_BMI = 0x400
FILTER_NOT_IN_FINAL_ASSESSMENTS = 0x800
FILTERP_ALL = 0xffffffff
patient_flag_descs = {
    0x1: 'other_territory',
    0x2: 'no_assessments', 0x4: 'one_assessment',
    0x8: 'missing_year_of_birth', 0x10: 'out_of_range_year_of_birth',
    0x20: 'missing_height', 0x40: 'out_of_range_height',
    0x80: 'missing_weight', 0x100: 'out_of_range_weight',
    0x200: 'missing_bmi', 0x400: 'out_of_range_bmi',
    0x800: 'not_in_final_assessments',
    0xffffffff: 'all_flags'
}

# assessment filter values
AFILTER_INVALID_PATIENT_ID = 0x1
AFILTER_PATIENT_FILTERED = 0x2
FILTER_MISSING_TEMP = 0x0
FILTER_BAD_TEMP = 0x8
FILTER_INCONSISTENT_NOT_TESTED = 0x10
FILTER_INCONSISTENT_TESTED = 0x20
FILTER_HEALTHY_BUT_SYMPTOMS = 0x40
FILTER_NOT_HEALTHY_BUT_NO_SYMPTOMS = 0x80
FILTER_INVALID_COVID_PROGRESSION = 0X100
FILTERA_ALL = 0xffffffff


assessment_flag_descs = {
    0x1: 'invalid_patient_id',
    0x2: 'patient_filtered',
    0x8: 'out_of_range_temperature',
    0x10: 'inconsistent_testing_not_tested', 0x20: 'inconsistent_testing_was_tested',
    0x40: 'inconsistent_symptoms', 0x80: 'inconsistent_no_symptoms',
    0x100: 'invalid_covid_progression',
    0xffffffff: 'all_flags'
}

symptomatic_fields = ["fatigue", "shortness_of_breath", "abdominal_pain", "chest_pain",
                      "delirium", "diarrhoea", "fever", "headache",
                      "hoarse_voice", "loss_of_smell", "persistent_cough",  "skipped_meals",
                      "sore_throat", "unusual_muscle_pains"]
flattened_fields = [("fatigue", "fatigue_binary"), ("shortness_of_breath", "shortness_of_breath_binary")]
exposure_fields = ["always_used_shortage", "have_used_PPE", "never_used_shortage", "sometimes_used_shortage",
                   "treated_patients_with_covid"]
miscellaneous_fields = ['health_status', 'location', 'level_of_isolation', 'had_covid_test', 'tested_covid_positive']
existing_fields = ('id', 'patient_id', 'created_at', 'updated_at', 'version',
                   'country_code', 'treatment', 'other_symptoms')
custom_field_aggregators = {
    'treatment': concatenate_maybe_strs,
    'other_symptoms': concatenate_maybe_strs
}


def pipeline(patient_filename, assessment_filename, data_schema, parsing_schema, year, territories=None):

    early_filter = None
    if territories is not None:
        territories = tuple(territories.split(','))
        early_filter = ('country_code', lambda x: x in territories)

    # categorical_maps = data_schema.assessment_categorical_maps
    # TODO: use proper logging throughout
    print(); print()
    print('load patients')
    print('=============')
    p_categorical_maps = {k: v for k, v in data_schema['patients'].fields.items()
                          if v.strings_to_values is not None}
    with open(patient_filename) as f:
        geoc_ds = dataset.Dataset(f, p_categorical_maps,
                                  early_filter=early_filter,
                                  show_progress_every=1000000,
                                  stop_after=500000)
    print("sorting patients")
    geoc_ds.sort(('id',))
    geoc_ds.show()
    print("patient row count:", geoc_ds.row_count())


    print(); print()
    print('load assessments')
    print('================')
    a_categorical_maps = {k: v for k, v in data_schema['assessments'].fields.items()
                          if v.strings_to_values is not None}
    with open(assessment_filename) as f:
        asmt_ds = dataset.Dataset(f, a_categorical_maps,
                                  early_filter=early_filter,
                                  show_progress_every=1000000,
                                  stop_after=15000000)
    print('sorting assessments')
    asmt_ds.sort(('patient_id', 'updated_at'))
    asmt_ds.show()
    print("assessment row count:", asmt_ds.row_count())


    print(); print()
    print("pre-sort by patient id")
    print("======================")

    print("pre-sort patient data")
    geoc_filter_status = np.zeros(geoc_ds.row_count(), dtype=np.uint32)

    print(); print("pre-sort assessment data")
    asmt_filter_status = np.zeros(asmt_ds.row_count(), dtype=np.uint32)

    # if territories is not None:
    #     print(); print();
    #     print("filter patients from outside the territory of interest")
    #     print("------------------------------------------------------")
    #
    #     country_codes = geoc_ds.field_by_name('country_code')
    #     for ir, r in enumerate(country_codes):
    #         if r != territories:
    #             geoc_filter_status[ir] |= PFILTER_OTHER_TERRITORY
    #     print(f'other territories: filtered {count_flag_set(geoc_filter_status, PFILTER_OTHER_TERRITORY)} missing values')

    print('patients:', len(geoc_filter_status))

    print(); print()
    print("patients")
    print("--------")

    ptnt_dest_fields = dict()

    print()
    print("checking age")
    src_yobs = geoc_ds.field_by_name('year_of_birth')
    ages = np.zeros(len(src_yobs), dtype=np.uint32)
    fn = CalculateAgeFromYearOfBirth(FILTER_MISSING_AGE, FILTER_BAD_AGE,
                                     valid_range_fac_inc(MIN_AGE, MAX_AGE), year)
    fn(src_yobs, ages, geoc_filter_status)
    ptnt_dest_fields['age'] = ages
    print(f'age: filtered {count_flag_set(geoc_filter_status, FILTER_MISSING_AGE)} missing values')
    print(f'age: filtered {count_flag_set(geoc_filter_status, FILTER_BAD_AGE)} bad values')


    print()
    print('checking weight / height / bmi')
    src_genders = geoc_ds.field_by_name('gender')
    src_weights = geoc_ds.field_by_name('weight_kg')
    src_heights = geoc_ds.field_by_name('height_cm')
    src_bmis = geoc_ds.field_by_name('bmi')

    fn_fac = parsing_schema.class_entries['validate_weight_height_bmi']
    fn = fn_fac(MIN_WEIGHT, MAX_WEIGHT, MIN_HEIGHT, MAX_HEIGHT, MIN_BMI, MAX_BMI,
                FILTER_MISSING_AGE, FILTER_BAD_AGE,
                FILTER_MISSING_WEIGHT, FILTER_BAD_WEIGHT,
                FILTER_MISSING_HEIGHT, FILTER_BAD_HEIGHT,
                FILTER_MISSING_BMI, FILTER_BAD_BMI)
    weight_clean, height_clean, bmi_clean =\
        fn(src_genders, ages, src_weights, src_heights, src_bmis, geoc_filter_status)
    ptnt_dest_fields['weight_clean'] = weight_clean
    ptnt_dest_fields['height_clean'] = height_clean
    ptnt_dest_fields['bmi_clean'] = bmi_clean
    print(f'weight: filtered {count_flag_set(geoc_filter_status, FILTER_MISSING_WEIGHT)} missing_values')
    print(f'weight: filtered {count_flag_set(geoc_filter_status, FILTER_BAD_WEIGHT)} missing_values')
    print(f'height: filtered {count_flag_set(geoc_filter_status, FILTER_MISSING_HEIGHT)} missing_values')
    print(f'height: filtered {count_flag_set(geoc_filter_status, FILTER_BAD_HEIGHT)} missing_values')
    print(f'bmi: filtered {count_flag_set(geoc_filter_status, FILTER_MISSING_BMI)} missing_values')
    print(f'bmi: filtered {count_flag_set(geoc_filter_status, FILTER_BAD_BMI)} missing_values')

    print(); print('unfiltered patients:', count_flag_empty(geoc_filter_status))

    #TODO: save patients here, including filter flags



    print(); print()
    print("assessments")
    print("-----------")

    asmt_dest_fields = dict()
    asmt_dest_keys = dict()

    patient_ids = set()
    src_patient_ids = geoc_ds.field_by_name('id')
    for ir, r in enumerate(src_patient_ids):
        if geoc_filter_status[ir] == 0:
            patient_ids.add(r)
    src_asmt_patient_ids = asmt_ds.field_by_name('patient_id')
    for ir, r in enumerate(src_asmt_patient_ids):
        if r not in patient_ids:
            asmt_filter_status[ir] |= AFILTER_PATIENT_FILTERED

    print('assessments filtered due to patient filtering:',
          count_flag_set(asmt_filter_status, AFILTER_PATIENT_FILTERED))
    print('assessments filtered total:',
          count_flag_set(asmt_filter_status, FILTERA_ALL))

    print(); print("checking temperature")
    fn_fac = parsing_schema.class_entries['validate_temperature']
    fn = fn_fac(MIN_TEMP, MAX_TEMP, FILTER_MISSING_TEMP, FILTER_BAD_TEMP)
    temperature_c = fn(asmt_ds.field_by_name('temperature'), asmt_filter_status)
    asmt_dest_fields['temperature_C'] = temperature_c
    print(f'temperature: filtered {count_flag_set(asmt_filter_status, FILTER_BAD_TEMP)} bad values')

    # print(); print("checking inconsistent test / test results fields")
    # src_had_test = asmt_ds.field_by_name('had_covid_test')
    # src_tested_covid_positive = asmt_ds.field_by_name('tested_covid_positive')
    # fn = CheckTestingConsistency(FILTER_INCONSISTENT_NOT_TESTED, FILTER_INCONSISTENT_TESTED)
    # fn(src_had_test, src_tested_covid_positive, asmt_filter_status)
    # print(f'inconsistent_not_tested: filtered {count_flag_set(asmt_filter_status, FILTER_INCONSISTENT_NOT_TESTED)} missing values')
    # print(f'inconsistent_tested: filtered {count_flag_set(asmt_filter_status, FILTER_INCONSISTENT_TESTED)} missing values')


    print(); print('unfiltered assessments:', np.count_nonzero(asmt_filter_status == 0))


    print(); print()
    print("convert symptomatic, exposure, flattened and miscellaneous fields to bool")
    print("-------------------------------------------------------------------------")

    any_symptoms = np.zeros(asmt_ds.row_count(), dtype=np.bool)
    for s in symptomatic_fields:
        print(f"symptomatic_field '{s}' to categorical")

        asmt_dest_fields[s] = copy_field(asmt_ds.field_by_name(s))
        any_symptoms |= asmt_dest_fields[s] > 1
        print(np.count_nonzero(asmt_dest_fields[s] == True))
        print(np.count_nonzero(any_symptoms == True))

    print(build_histogram(asmt_ds.field_by_name('tested_covid_positive')))

    for f in flattened_fields:
        flattened = {'': 0, 'no': 1, 'mild': 2, 'significant': 2, 'severe': 2}
        print(f"flattened_field '{f[0]}' to categorical field '{f[1]}'")
        remap = map_between_categories(a_categorical_maps[f[0]].strings_to_values,
                                       flattened)
                                       # a_categorical_maps[f[1]].strings_to_values)
        asmt_dest_fields[f[1]] =\
            to_categorical(asmt_ds.field_by_name(f[0]), remap)
        # TODO: this shouldn't be necessary as the fields were covered in 'symptomatic_fields'
        any_symptoms |= asmt_dest_fields[f[1]] > 1

    for e in exposure_fields:
        print(f"exposure_field '{e}' to categorical")
        asmt_dest_fields[e] = copy_field(asmt_ds.field_by_name(e))
    for m in miscellaneous_fields:
        print(f"miscellaneous_field '{m}' to categorical")
        asmt_dest_fields[m] = copy_field(asmt_ds.field_by_name(m))


    print(); print()
    print("validate health status with symptoms")
    print("---------------------------------")
    fn = CheckInconsistentSymptoms(FILTER_HEALTHY_BUT_SYMPTOMS, FILTER_NOT_HEALTHY_BUT_NO_SYMPTOMS)
    # TODO: keys should be got from the dataset once it is loaded rather than referring to the categorical maps directly
    fn(asmt_dest_fields['health_status'], any_symptoms, asmt_filter_status,
       a_categorical_maps['health_status'].strings_to_values['healthy'],
       a_categorical_maps['health_status'].strings_to_values['not_healthy'])
    for f in (FILTER_HEALTHY_BUT_SYMPTOMS, FILTER_NOT_HEALTHY_BUT_NO_SYMPTOMS):
        print(f'{assessment_flag_descs[f]}: {count_flag_set(asmt_filter_status, f)}')

    print(); print('unfiltered assessments:', np.count_nonzero(asmt_filter_status == 0))


    # validate assessments per patient
    print(); print()
    print("validate covid progression")
    print("--------------------------")
    sanitised_hct_covid_results = np.ndarray(asmt_ds.row_count(), dtype=np.uint8)
    sanitised_covid_results = np.ndarray(asmt_ds.row_count(), dtype=np.uint8)
    sanitised_covid_results_key = \
        list(a_categorical_maps['tested_covid_positive'].values_to_strings.values())

    fn_fac = parsing_schema.class_entries['clean_covid_progression']
    fn = fn_fac(asmt_ds.field_by_name('had_covid_test'), asmt_ds.field_by_name('tested_covid_positive'),
                asmt_filter_status,
                sanitised_hct_covid_results, sanitised_covid_results,
                FILTER_INVALID_COVID_PROGRESSION)
    iterate_over_patient_assessments2(
        asmt_ds.field_by_name('patient_id'), asmt_filter_status, fn)

    print(f'{assessment_flag_descs[FILTER_INVALID_COVID_PROGRESSION]}:',
          count_flag_set(asmt_filter_status, FILTER_INVALID_COVID_PROGRESSION))

    asmt_dest_fields['tested_covid_positive_clean'] = sanitised_covid_results
    asmt_dest_keys['tested_covid_positive_clean'] = sanitised_covid_results_key
    asmt_dest_fields['had_covid_test_clean'] = sanitised_hct_covid_results
    asmt_dest_keys['had_covid_test_clean'] = sanitised_covid_results_key

    print(); print("checking inconsistent test / test results fields")
    fn = CheckTestingConsistency(FILTER_INCONSISTENT_NOT_TESTED, FILTER_INCONSISTENT_TESTED)
    fn(sanitised_hct_covid_results, sanitised_covid_results, asmt_filter_status)
    print(f'inconsistent_not_tested: filtered {count_flag_set(asmt_filter_status, FILTER_INCONSISTENT_NOT_TESTED)} missing values')
    print(f'inconsistent_tested: filtered {count_flag_set(asmt_filter_status, FILTER_INCONSISTENT_TESTED)} missing values')

    print('remaining assessments before squashing', np.count_nonzero(asmt_filter_status == 0))

    # create a new assessment space with only unfiltered rows
    print(); print()
    print("discard all filtered assessments")
    print("--------------------------------")
    remaining_asmt_fields = list()
    remaining_dest_fields = dict()

    filter_map = list()
    for ir, r in enumerate(asmt_filter_status):
        if r == 0:
            filter_map.append(ir)

    for ir, r in enumerate(asmt_ds.fields_):
        remaining_asmt_fields.append(filtered_field.FilteredField(r, filter_map))

    for k, v in asmt_dest_fields.items():
        remaining_dest_fields[k] = filtered_field.FilteredField(v, filter_map)

    print("remaining asmt fields: ", len(filter_map))
    remaining_asmt_filter_status = [0] * len(filter_map)

    print(); print()
    print("quantise assessments by day")
    print("---------------------------")

    fn = CalculateMergedFieldCount(remaining_asmt_fields[asmt_ds.field_to_index('updated_at')])
    print(len(filter_map))
    print(len(remaining_asmt_filter_status))
    remaining_patient_ids = remaining_asmt_fields[asmt_ds.field_to_index('patient_id')]
    iterate_over_patient_assessments2(remaining_patient_ids, remaining_asmt_filter_status, fn)
    remaining_asmt_row_count = len(filter_map) - fn.merged_row_count
    print(f'{len(filter_map)} - {fn.merged_row_count} = {remaining_asmt_row_count}')

    existing_field_indices = [(f, asmt_ds.field_to_index(f)) for f in existing_fields]

    resulting_fields = dict()
    for e in existing_fields:
        resulting_fields[e] = [None] * remaining_asmt_row_count
    for dk, dv in remaining_dest_fields.items():
        resulting_fields[dk] = np.zeros((remaining_asmt_row_count, ), dtype=dv.dtype)

    resulting_field_keys = dict()
    for dk, dv in asmt_dest_keys.items():
        resulting_field_keys[dk] = dv

    resulting_filter_status = np.zeros(remaining_asmt_row_count, dtype=np.uint32)

    print('remaining_dest_len:', len(remaining_dest_fields['fatigue_binary']))
    print('resulting_fields:', len(resulting_fields['patient_id']))

    print(build_histogram(remaining_dest_fields['tested_covid_positive_clean']))
    concat_field_indices =\
        [asmt_ds.field_to_index('other_symptoms'), asmt_ds.field_to_index('treatment')]
    merge = MergeAssessmentRows(concat_field_indices,
                                resulting_fields, remaining_dest_fields,
                                existing_field_indices, custom_field_aggregators,
                                remaining_asmt_filter_status, resulting_filter_status)
    iterate_over_patient_assessments(remaining_asmt_fields, remaining_asmt_filter_status, merge)
    print(merge.rfindex)

    unique_patients = defaultdict(int)
    for r in remaining_patient_ids:
        unique_patients[r] += 1
    print('unique patents in remaining assessments:', len(unique_patients))

    print(); print()
    print("filter summaries")
    print("----------------")

    print(); print('patient flags set')
    for v in patient_flag_descs.keys():
        print(f'{patient_flag_descs[v]}: {count_flag_set(geoc_filter_status, v)}')

    print(); print('assessment flags set')
    for v in assessment_flag_descs.keys():
        print(f'{assessment_flag_descs[v]}: {count_flag_set(asmt_filter_status, v)}')

    return (geoc_ds, geoc_filter_status, ptnt_dest_fields,
            asmt_ds, asmt_filter_status,
            remaining_asmt_fields, remaining_asmt_filter_status,
            resulting_fields, resulting_field_keys)


def regression_test_assessments(old_assessments, new_assessments):

    with open(old_assessments) as f:
        r_a_ds = dataset.Dataset(f)
    r_a_ds.sort(('patient_id', 'id'))
    with open(new_assessments) as f:
        p_a_ds = dataset.Dataset(f)
    p_a_ds.sort(('patient_id', 'id'))

    # r_a_fields = r_a_ds.fields_
    # p_a_fields = p_a_ds.fields_

    r_a_keys = set(r_a_ds.names_)
    p_a_keys = set(p_a_ds.names_)
    print(r_a_keys)
    print(p_a_keys)
    print('keys only in r:', r_a_keys.difference(p_a_keys))
    print('keys only in p:', p_a_keys.difference(r_a_keys))

    diagnostic_row_keys = ['id', 'patient_id', 'created_at', 'updated_at', 'health_status', 'fatigue', 'fatigue_binary', 'had_covid_test', 'tested_covid_positive']
    comparison_keys = diagnostic_row_keys
    r_fns = {'created_at': regression.datetime_compare_to_secs,
             'updated_at': regression.datetime_compare_to_secs}

    patients_with_disparities = set()
    r = 0
    p = 0
    r_ids = r_a_ds.field_by_name('id')
    r_pids = r_a_ds.field_by_name('patient_id')
    r_upda = r_a_ds.field_by_name('day')
    r_hct = r_a_ds.field_by_name('had_covid_test')
    r_tcp = r_a_ds.field_by_name('tested_covid_positive')

    p_ids = p_a_ds.field_by_name('id')
    p_pids = p_a_ds.field_by_name('patient_id')
    p_upda = p_a_ds.field_by_name('day')
    p_hct = p_a_ds.field_by_name('had_covid_test')
    p_tcp = p_a_ds.field_by_name('tested_covid_positive')
    while r < r_a_ds.row_count() and p < p_a_ds.row_count():
        rkey = (r_pids[r], r_upda[r])
        pkey = (p_pids[p], p_upda[p])
        if rkey < pkey:
            print(f'{r}, {p}: {rkey} not in python dataset')
            print_diagnostic_row('', r_a_ds, r, diagnostic_row_keys, fns=r_fns)
            print_diagnostic_row('', p_a_ds, p, diagnostic_row_keys)
            patients_with_disparities.add(r_pids[r])
            r += 1
        elif pkey < rkey:
            print(f'{r}, {p}: {pkey} not in r dataset')
            print_diagnostic_row('', r_a_ds, r, diagnostic_row_keys, fns=r_fns)
            print_diagnostic_row('', p_a_ds, p, diagnostic_row_keys)
            patients_with_disparities.add(p_pids[p])
            p += 1
        else:
            disparities = regression.check_row(r_a_ds, r, p_a_ds, p, comparison_keys, r_fns)
            if disparities != None:
                print(p_ids[p], ','.join(disparities))
            r += 1
            p += 1

        # if r < r_a_ds.row_count():
        #     treatment = r_a_fields[r_a_ds.field_to_index('treatment')][r]
        #     if treatment not in ('NA', '', 'none'):
        #         print(r, treatment)


    r_a_ds.sort(('patient_id', 'updated_at'))
    p_a_ds.sort(('patient_id', 'updated_at'))

    for pd in patients_with_disparities:
        print(); print(pd)
        for ir in range(r_a_ds.row_count()):
            if r_a_ds.value_from_fieldname(ir, 'patient_id') == pd:
                print_diagnostic_row(f'r[ir]', r_a_ds, ir, diagnostic_row_keys, fns=r_fns)
        for ip in range(p_a_ds.row_count()):
            if p_a_ds.value_from_fieldname(ir, 'patient_id') == pd:
                print_diagnostic_row(f'p[ip]', p_a_ds, ip, diagnostic_row_keys)

    print('done')


def regression_test_patients(old_patients, new_patients):
    print(); print('regression test patients')
    print('old_patients:', old_patients)
    print('new_patients:', new_patients)
    with open(old_patients) as f:
        r_a_ds = dataset.Dataset(f)
    r_a_ds.sort(('id',))
    with open(new_patients) as f:
        p_a_ds = dataset.Dataset(f)
    p_a_ds.sort(('id',))

    r_a_keys = set(r_a_ds.names_)
    p_a_keys = set(p_a_ds.names_)
    print('r_a_keys:', r_a_keys)
    print('p_a_keys:', p_a_keys)
    print('keys only in r:', r_a_keys.difference(p_a_keys))
    print('keys only in p:', p_a_keys.difference(r_a_keys))

    print('checking for disparities')
    patients_with_disparities = set()
    r = 0
    p = 0
    r_ids = r_a_ds.field_by_name('id')
    r_ages = r_a_ds.field_by_name('age')
    r_weight = r_a_ds.field_by_name('weight_kg')
    r_height = r_a_ds.field_by_name('height_cm')
    r_bmi = r_a_ds.field_by_name('bmi')
    p_ids = p_a_ds.field_by_name('id')
    p_ages = p_a_ds.field_by_name('age')
    p_weight = p_a_ds.field_by_name('weight_clean')
    p_height = p_a_ds.field_by_name('height_clean')
    p_bmi = p_a_ds.field_by_name('bmi_clean')
    while r < r_a_ds.row_count() and p < p_a_ds.row_count():
        rkey = r_ids[r]
        pkey = p_ids[p]
        if rkey < pkey:
            print(f'{r}, {p}: {rkey} not in python dataset')
            patients_with_disparities.add(rkey)
            r += 1
        elif pkey < rkey:
            print(f'{r}, {p}: {pkey} not in r dataset')
            patients_with_disparities.add(pkey)
            p += 1
        else:
            age_same = r_ages[r] == p_ages[p]
            weight_same = abs(float(r_weight[r]) - float(p_weight[p])) < 0.00001
            height_same = abs(float(r_height[r]) - float(p_height[p])) < 0.00001
            bmi_same = abs(float(r_bmi[r]) - float(p_bmi[p])) < 0.00001
            if not age_same or not weight_same or not height_same or not bmi_same:
                print(r, p,
                      r_ids[r], p_ids[p],
                      'na' if r_ages[r] is '' else r_ages[r], p_ages[p] if '' else p_ages[p],
                      'na' if r_weight[r] is '' else r_weight[r], 'na' if p_weight[p] is '' else p_weight[p],
                      'na' if r_height[r] is '' else r_height[r], 'na' if p_height[p] is '' else p_height[p],
                      'na' if r_bmi[r] is '' else r_bmi[r], 'na' if p_bmi[p] is '' else p_bmi[p]
                      )
            r += 1
            p += 1

    for pd in patients_with_disparities:
        print(); print(pd)
    print('checking_for_disparities: done')


def save_csv(pipeline_output, patient_data_out, assessment_data_out, data_schema):
    a_categorical_maps = \
        {k: v for k, v in data_schema["assessments"].fields.items()
         if v.strings_to_values is not None and v.out_of_range_label is None}

    p_ds, p_status, p_dest_fields,\
    a_ds, a_status, ra_fields, ra_status, res_fields, res_keys \
        = pipeline_output
    remaining_patients = set()
    for p in res_fields['patient_id']:
        remaining_patients.add(p)
    for ip, p in enumerate(p_ds.field_by_name('id')):
        if p not in remaining_patients:
            p_status[ip] |= FILTER_NOT_IN_FINAL_ASSESSMENTS

    print()
    print(f'writing patient data to {patient_data_out}')
    tstart = time.time()
    with open(patient_data_out, 'w') as f:
        dest_keys = list(p_dest_fields.keys())
        values = [None] * (len(p_ds.names_) + len(dest_keys))
        csvw = csv.writer(f)
        csvw.writerow(p_ds.names_ + dest_keys)
        for ir in range(p_ds.row_count()):
            if p_status[ir] == 0:
                for iv, v in enumerate(p_ds.fields_):
                    values[iv] = v[ir]
                for iv in range(len(dest_keys)):
                    values[len(p_ds.names_) + iv] = p_dest_fields[dest_keys[iv]][ir]
                csvw.writerow(values)
    print(f'written to {patient_data_out} in {time.time() - tstart} seconds')

    functor_fields = {'created_at': datetime_to_seconds, 'updated_at': datetime_to_seconds}

    print(f'writing assessment data to {assessment_data_out}')
    tstart = time.time()
    with open(assessment_data_out, 'w') as f:
        csvw = csv.writer(f)
        headers = list(res_fields.keys())
        # TODO: constructed fields should be in their own collection; the ['day'] and +1 stuff is a temporary hack
        csvw.writerow(headers + ['day'])
        row_field_count = len(res_fields)
        row_values = [None] * (row_field_count + 1)
        for ir in range(len(res_fields['id'])):
            if ra_status[ir] == 0:
                for irh, rh in enumerate(headers):
                    if rh in functor_fields:
                        row_values[irh] = functor_fields[rh](res_fields[rh][ir])
                    elif rh in a_categorical_maps:
                        v_to_s = a_categorical_maps[rh].values_to_strings
                        if res_fields[rh][ir] >= len(v_to_s):
                            print(f'{res_fields[rh][ir]} is out of range for {v_to_s}')
                        row_values[irh] = v_to_s[res_fields[rh][ir]]
                    else:
                        row_values[irh] = res_fields[rh][ir]
                updated = res_fields['updated_at']
                row_values[-1] = f"{updated[ir][0:4]}-{updated[ir][5:7]}-{updated[ir][8:10]}"
                csvw.writerow(row_values)
                for irv in range(len(row_values)):
                    row_values[irv] = None
    print(f'written to {assessment_data_out} in {time.time() - tstart} seconds')


def save_patient_npys(pipeline_output,
                      patient_data_out, assessment_data_out, test_data_out,
                      data_schema):

    categorical_maps = data_schema.assessment_categorical_maps

    p_ds, p_status, p_dest_fields,\
    a_ds, a_status, ra_fields, ra_status, res_fields, res_keys \
        = pipeline_output


# TODO: add json based config option
# TODO: add parsing schema option and default
# TODO: add data schema option and default
# TODO: provide specific schema element overrides (via json only)
# TODO: add flag to output values as categorical variables
if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version='v0.1.9')
    parser.add_argument('-r', '--regression_test', action='store_true')
    parser.add_argument('-te', '--territories', default=None,
                        help='the territories to filter the dataset on (runs on all territories if not set)')
    parser.add_argument('-p', '--patient_data',
                        help='the location and name of the patient data csv file')
    parser.add_argument('-po', '--patient_data_out',
                        help='the location and name of the output patient data csv file')
    parser.add_argument('-a', '--assessment_data',
                        help='the location and name of the assessment data csv file')
    parser.add_argument('-ao', '--assessment_data_out',
                        help='the location and name of the output assessment data csv file')
    parser.add_argument('-ps', '--parsing_schema', default=1, type=int,
                        help='the schema number to use for parsing and cleaning data')
    parser.add_argument('-y', '--year', default=datetime.datetime.now().year, type=int)
    args = parser.parse_args()

    if args.parsing_schema not in parsing_schemas.parsing_schemas:
        error_str = "the parsing schema must be one of {} for this version"
        print(error_str.format(parsing_schemas.parsing_schemas))
        exit(-1)

    if args.regression_test:
        regression_test_assessments('assessments_cleaned_short.csv', args.assessment_data)
        regression_test_patients('patients_cleaned_short.csv', args.patient_data)
    else:
        tstart = time.time()

        data_schema_version = 1
        early_filter = None
        parsing_schema_version = args.parsing_schema
        parsing_schema = parsing_schemas.ParsingSchema(parsing_schema_version)
        pipeline_output = pipeline(args.patient_data, args.assessment_data,
                                   parsing_schema, args.year,
                                   territories=args.territories)
        print(f'cleaning completed in {time.time() - tstart} seconds')

        save_csv(pipeline_output, args.patient_data_out, args.assessment_data_out)



import math
from datetime import datetime
from collections import defaultdict
import numpy as np
import h5py
import pandas as pd

from hystore.core import exporter, persistence, utils
from hystore.core.persistence import DataStore
from hystore.processing.nat_medicine_model import nature_medicine_model_1


def method_paper_prediction_pipeline(ds, src_data, dest_data, first_timestamp, last_timestamp):
    s_asmts = src_data['assessments']
    s_tests = src_data['tests']
    flat_asmts = dest_data.create_group('flat_asmts')
    flat_tests = dest_data.create_group('flat_tests')

    first_dt = datetime.fromtimestamp(first_timestamp)
    last_dt = datetime.fromtimestamp(last_timestamp)
    print(s_tests.keys())

    # Filter tests
    # ============

    t_cats = ds.get_reader(s_tests['created_at'])
    raw_t_cats = t_cats[:]
    t_dts = ds.get_reader(s_tests['date_taken_specific'])
    raw_t_dts = t_dts[:]
    t_dsbs = ds.get_reader(s_tests['date_taken_between_start'])
    raw_t_dsbs = t_dsbs[:]
    t_dsbe = ds.get_reader(s_tests['date_taken_between_end'])
    raw_t_dsbe = t_dsbe[:]

    # remove non GB tests
    cur_filter = (ds.get_reader(s_tests['country_code'])[:] == b'GB')
    test_filter = cur_filter[:]

    # remove tests where no dates are set
    cur_filter = np.logical_not((raw_t_dts == 0) & (raw_t_dsbs == 0) & (raw_t_dsbe == 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 1:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where all three dates are set
    cur_filter = np.logical_not((raw_t_dts != 0) & (raw_t_dsbs != 0) & (raw_t_dsbe != 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 2:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where only one of the date range tests is set
    cur_filter = np.logical_not((raw_t_dsbs != 0) & (raw_t_dsbe == 0) |
                                (raw_t_dsbs == 0) & (raw_t_dsbe != 0))
    test_filter = test_filter & cur_filter
    print("standard test filter 3:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where specific date is set but out of range
    cur_filter =\
        (raw_t_dts == 0) | ((raw_t_dts >= first_timestamp) & (raw_t_dts <= last_timestamp))
    test_filter = test_filter & cur_filter
    print("standard test filter 4:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where beginning date is set but out of range
    cur_filter =\
        (raw_t_dsbs == 0) | ((raw_t_dsbs >= first_timestamp) & (raw_t_dsbs <= last_timestamp))
    test_filter = test_filter & cur_filter
    print("standard test filter 5:", np.count_nonzero(test_filter), len(test_filter))

    # remove tests where ending date is set but out of range
    cur_filter = \
        (raw_t_dsbe == 0) | ((raw_t_dsbe >= first_timestamp) & (raw_t_dsbe <= last_timestamp))
    test_filter = test_filter & cur_filter
    print("standard test filter 6:", np.count_nonzero(test_filter), len(test_filter))

    test_timestamps = np.where(raw_t_dts != 0,
                               raw_t_dts,
                               raw_t_dsbs + (raw_t_dsbe - raw_t_dsbs) / 2)

    # remove tests where the test date is after the created at date
    cur_filter = test_timestamps <= raw_t_cats
    test_filter = test_filter & cur_filter
    print("standard test filter 7:", np.count_nonzero(test_filter), len(test_filter))

    t_rsts = ds.get_reader(s_tests['result'])
    t_rsts.get_writer(flat_tests, 'result').write(ds.apply_filter(test_filter, t_rsts))
    t_pids = ds.get_reader(s_tests['patient_id'])
    t_pids.get_writer(flat_tests, 'patient_id').write(ds.apply_filter(test_filter, t_pids))
    ds.get_timestamp_writer(flat_tests, 'eff_test_date').write(
        ds.apply_filter(test_filter, test_timestamps))

    # test_min_ts = datetime.fromtimestamp(test_timestamps[test_filter].min())
    # test_max_ts = datetime.fromtimestamp(test_timestamps[test_filter].max())
    # print(test_min_ts, test_max_ts)

    # Filter assessments
    # ------------------

    symptoms = ('persistent_cough', 'fatigue', 'delirium', 'shortness_of_breath', 'fever',
                'diarrhoea', 'abdominal_pain', 'chest_pain', 'hoarse_voice', 'skipped_meals',
                'loss_of_smell')
    symptom_thresholds = {s: 2 for s in symptoms}
    symptom_thresholds['fatigue'] = 3
    symptom_thresholds['shortness_of_breath'] = 3

    with utils.Timer("filter all out of date range assessments and non-uk assessments"):
        a_cats = ds.get_reader(s_asmts['created_at'])[:]
        # in_date_range = (a_cats >= first_timestamp) & (a_cats < last_timestamp)
        in_date_range = a_cats >= first_timestamp
        in_date_range = in_date_range & (ds.get_reader(s_asmts['country_code'])[:] == b'GB')

    with utils.Timer("get indices of final assessments of each day"):
        reader = ds.apply_filter(in_date_range, ds.get_reader(s_asmts['created_at_day'])[:])
        spans = ds.get_spans(reader)
        last_daily_asmts = np.zeros(len(spans)-1, dtype='int64')
        ds.apply_spans_index_of_last(spans, last_daily_asmts)
        print(len(last_daily_asmts))

    pc = ds.get_reader(s_asmts['persistent_cough'])[:]
    pc1 = ds.apply_indices(last_daily_asmts, ds.apply_filter(in_date_range, pc))
    pc2 = ds.apply_indices(last_daily_asmts, pc)
    print(len(pc1), len(pc2))
    print(np.array_equal(pc1, pc2))


    with utils.Timer("flattening and filtering symptoms"):
        for s in symptoms:
            reader = ds.get_reader(s_asmts[s])
            writer = ds.get_numeric_writer(flat_asmts, s, 'bool')
            filtered = ds.apply_indices(last_daily_asmts, ds.apply_filter(in_date_range, reader[:]))
            writer.write(filtered >= symptom_thresholds[s])

    with utils.Timer("flattening and filtering other fields", new_line=True):
        for f in ('id', 'patient_id', 'created_at', 'created_at_day', 'tested_covid_positive'):
            reader = ds.get_reader(s_asmts[f])
            writer = reader.get_writer(flat_asmts, f)
            ds.apply_filter(in_date_range, reader, writer)
            reader = ds.get_reader(flat_asmts[f])
            writer = reader.get_writer(flat_asmts, f, write_mode='overwrite')
            ds.apply_indices(last_daily_asmts, reader, writer)
            print("  {}".format(f), len(ds.get_reader(flat_asmts[f])))

    # telemetry only
    for s in symptoms:
        print(s, len(ds.get_reader(flat_asmts[s])),
              np.count_nonzero(ds.get_reader(flat_asmts[s])[:]))


    # Filter tests
    # ------------

    # # filter tests within day range first
    # t_cats = ds.get_reader(s_tests['created_at'])
    # raw_t_cats = t_cats[:]
    # t_rsts = ds.get_reader(s_tests['result'])
    # t_pids = ds.get_reader(s_tests['patient_id'])
    # # test_date_filter = (raw_t_cats >= first_timestamp) & (raw_t_cats < last_timestamp)
    # test_date_filter = raw_t_cats >= first_timestamp
    # test_date_filter = test_date_filter & (ds.get_reader(s_tests['country_code'])[:] == b'GB')
    # t_cats.get_writer(flat_tests, 'created_at').write(ds.apply_filter(test_date_filter, raw_t_cats))
    # t_rsts.get_writer(flat_tests, 'result').write(ds.apply_filter(test_date_filter, t_rsts))
    # t_pids.get_writer(flat_tests, 'patient_id').write(ds.apply_filter(test_date_filter, t_pids))
    #
    # raw_t_cats = ds.get_reader(flat_tests['created_at'])[:]
    # min_test_day = datetime.fromtimestamp(np.min(raw_t_cats))
    # max_test_day = datetime.fromtimestamp(np.max(raw_t_cats))
    # print(min_test_day, max_test_day)

    # Calculate prevalence
    # --------------------

    intercept = -1.19015973
    weights = {'persistent_cough': 0.23186655,
               'fatigue': 0.56532346,
               'delirium': -0.12935112,
               'shortness_of_breath': 0.58273967,
               'fever': 0.16580974,
               'diarrhoea': 0.10236126,
               'abdominal_pain': -0.11204163,
               'chest_pain': -0.12318634,
               'hoarse_voice': -0.17818597,
               'skipped_meals': 0.25902482,
               'loss_of_smell': 1.82895239}

    with utils.Timer("predicting covid by assessment", new_line=True):
        cumulative = np.zeros(len(ds.get_reader(flat_asmts['persistent_cough'])), dtype='float64')
        for s in symptoms:
            reader = ds.get_reader(flat_asmts[s])
            cumulative += reader[:] * weights[s]
        cumulative += intercept
        print("  {}".format(len(cumulative)))
        ds.get_numeric_writer(flat_asmts, 'prediction', 'float32').write(cumulative)
        pos_filter = cumulative > 0.0
        print("pos_filter: ", np.count_nonzero(pos_filter), len(pos_filter))


    # apply
    # positive test -> imputed positive -> negative test
    spans = ds.get_spans(ds.get_reader(flat_asmts['patient_id'])[:])
    print('spans:', len(spans))

    # generate a numpy array for each day, where each entry in the array is a patient with
    # assessments still in the dataset after the initial filter
    timedelta = datetime.fromtimestamp(last_timestamp) - datetime.fromtimestamp(first_timestamp)
    # print(datetime.fromtimestamp(first_timestamp))
    # print(datetime.fromtimestamp(last_timestamp))
    # print(timedelta)

    daydict = defaultdict(int)
    with utils.Timer("checking date deltas", new_line=True):
        a_cats = ds.get_reader(flat_asmts['created_at'])[:]
        first_day = datetime.fromtimestamp(first_timestamp)
        for i_r in range(len(a_cats)):
            daydict[(datetime.fromtimestamp(a_cats[i_r]) - first_day).days] += 1
        sdaydict = sorted(daydict.items())
        # for i_s in range(len(sdaydict)):
        #     print("sdaydict:", sdaydict[i_s][0], sdaydict[i_s][1])
        # print(sum(daydict.values()))

    # build a combined id index for assessments and tests
    remaining_a_pids = ds.get_reader(flat_asmts['patient_id'])[:]
    remaining_t_pids = ds.get_reader(flat_tests['patient_id'])[:]
    print("pids from assessments and tests:", len(remaining_a_pids), len(remaining_t_pids),
          len(set(remaining_a_pids).union(set(remaining_t_pids))))
    a_pid_index, t_pid_index = ds.get_shared_index((remaining_a_pids, remaining_t_pids))
    print("merging indices:", len(a_pid_index), len(t_pid_index), max(np.max(a_pid_index), np.max(t_pid_index)))

    max_index = max(a_pid_index[-1], t_pid_index[-1])
    print('max indices:', a_pid_index[-1], t_pid_index[-1])
    i_days = list([None] * len(sdaydict))
    t_days = list([None] * len(sdaydict))
    for i in range(len(sdaydict)):
        i_days[i] = np.zeros(max_index+1, dtype='int8')
        t_days[i] = np.zeros(max_index+1, dtype='int8')

    first_day = datetime.fromtimestamp(first_timestamp)
    a_cats = ds.get_reader(flat_asmts['created_at'])[:]
    a_tcps = ds.get_reader(flat_asmts['tested_covid_positive'])[:]
    a_offset_days = np.zeros(len(a_cats), dtype='int8')
    print(type((datetime.fromtimestamp(a_cats[0]) - first_day).days))

    # calculate offset days for assessments
    for i_r, r in enumerate(a_cats):
        a_offset_days[i_r] = (datetime.fromtimestamp(a_cats[i_r]) - first_day).days
    print("assessment_dates:", sorted(utils.build_histogram(a_offset_days)))

    # incorporate assessment predictions and positive test results
    # note: a_offset_days is in assessment space
    with utils.Timer("incorporating assessments and assessment-based tests"):
        for i_r, r in enumerate(a_offset_days):
            # i_days[a_offset_days[i_r]][a_pid_index[i_r]] =\
            #     from_tcp if from_tcp != 0 else from_prediction
            from_prediction = 7 if cumulative[i_r] > 0.0 else -7
            i_days[a_offset_days[i_r]][a_pid_index[i_r]] = from_prediction
            from_tcp = 7 if a_tcps[i_r] == 3 else -7 if a_tcps[i_r] == 2 else 0
            t_days[a_offset_days[i_r]][a_pid_index[i_r]] = from_tcp

    # incorporate test results
    # ------------------------

    t_etds = ds.get_reader(flat_tests['eff_test_date'])
    raw_t_etds = t_etds[:]
    t_rsts = ds.get_reader(flat_tests['result'])
    t_pids = ds.get_reader(flat_tests['patient_id'])

    # calculate offset days for tests
    t_offset_days = np.zeros(len(raw_t_etds), dtype='int8')
    t_offset_dates = [None] * len(raw_t_etds)
    for i_r, r in enumerate(raw_t_etds):
        t_offset_days[i_r] = (datetime.fromtimestamp(raw_t_etds[i_r]) - first_day).days
        t_offset_dates[i_r] = datetime.fromtimestamp(raw_t_etds[i_r]).date()
    print("test_dates:", sorted(utils.build_histogram(t_offset_days)))
    print("test_dates2:", sorted(utils.build_histogram(t_offset_dates)))


    # incorporate test results by to appropriate day's entry
    with utils.Timer("incorprating test_results"):
        for i_r, r in enumerate(t_offset_days):
            day = t_days[t_offset_days[i_r]]
            if t_rsts[i_r] == 4:
                day[t_pid_index[i_r]] = 7
            elif t_rsts[i_r] == 3:
                day[t_pid_index[i_r]] = -7
            # day[t_pid_index[i_r]] = 7 if t_rsts[i_r] == 4 else -7 if t_rsts[i_r] == 3 else 0
            # if day[t_pid_index[i_r]] == 0:
            #     day[t_pid_index[i_r]] = 7 if t_rsts[i_r] == 4 else -7 if t_rsts[i_r] == 3 else 0
            # else:
            #     day[t_pid_index[i_r]] = 7 if t_rsts[i_r] == 4 else max(day[t_pid_index[i_r]], -7)

    for i_d, d in enumerate(i_days):
        print(i_d, np.count_nonzero(d))

    with utils.Timer("calculating progression"):
        for da in (i_days, t_days):
            for i_d in range(len(da)-1):
                prior_d = da[i_d]
                next_d = da[i_d + 1]
                next_d[:] = np.where(next_d != 0,
                                     next_d,
                                     np.where(prior_d > 0, prior_d-1, np.minimum(prior_d+1, 0)))
    for d in range(len(i_days)):
        i_d = i_days[d]
        t_d = t_days[d]
        i_present = np.count_nonzero(i_d != 0)
        i_positive = np.count_nonzero(i_d > 0)
        t_present = np.count_nonzero(t_d != 0)
        t_positive = np.count_nonzero(t_d > 0)
        c_d = np.where(t_d == 0, i_d, t_d)
        c_present = np.count_nonzero(c_d != 0)
        c_positive = np.count_nonzero(c_d > 0)

        print(d, i_present, i_positive, t_present, t_positive, c_present, c_positive,
              c_positive / c_present)


if __name__ == '__main__':
    datastore = DataStore()
    src_file = '/home/ben/covid/ds_20200731_full.hdf5'
    dest_file = '/home/ben/covid/ds_20200731_supplements.hdf5'
    with h5py.File(src_file, 'r+') as src_data:
        with h5py.File(dest_file, 'w') as dest_data:
            method_paper_prediction_pipeline(datastore, src_data, dest_data,
                                             datetime.strptime("2020-04-01", '%Y-%m-%d').timestamp(),
                                             datetime.strptime("2020-07-31", '%Y-%m-%d').timestamp())



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from hystore.core.utils import check_input_lengths
from hystore.core import persistence


class CheckInconsistentSymptoms:
    def __init__(self, f_healthy_but_symptoms, f_not_healthy_but_no_symptoms):
        self.f_healthy_but_symptoms = f_healthy_but_symptoms
        self.f_not_healthy_but_no_symptoms = f_not_healthy_but_no_symptoms

    def __call__(self, healthy, symptoms, flags, i_healthy, i_not_healthy):
        check_input_lengths(('healthy', 'symptoms', 'flags'), (healthy, symptoms, flags))
        for i_r in range(len(healthy)):
            if healthy[i_r] == i_healthy and symptoms[i_r]:
                flags[i_r] |= self.f_healthy_but_symptoms
            elif healthy[i_r] == i_not_healthy and not symptoms[i_r]:
                flags[i_r] |= self.f_not_healthy_but_no_symptoms


def check_inconsistent_symptoms_1(datastore, src_assessments, dest_assessments, timestamp=None):
    if timestamp is None:
        timestamp = datastore.timestamp

    generated_health_fields = ()
    # generated_health_fields = ('has_temperature',)

    # TODO: check that all fields are leaky booleans
    health_check_fields = (
        'fever', 'persistent_cough', 'fatigue', 'shortness_of_breath', 'diarrhoea',
        'delirium', 'skipped_meals', 'abdominal_pain', 'chest_pain', 'hoarse_voice',
        'loss_of_smell', 'headache', 'chills_or_shivers', 'eye_soreness', 'nausea',
        'dizzy_light_headed', 'red_welts_on_face_or_lips', 'blisters_on_feet', 'sore_throat',
        'unusual_muscle_pains'
    )

    health_status = datastore.get_reader(src_assessments['health_status'])
    health_status_array = health_status[:]

    combined_results = np.zeros(len(health_status), dtype=np.bool)

    for h in health_check_fields:
        if h not in src_assessments.keys():
            print(f"warning: field {h} is not present in this dataset")
        else:
            f = datastore.get_reader(src_assessments[h])
            combined_results = combined_results & (f[:] != 2)

    for h in generated_health_fields:
        if h not in src_assessments.keys():
            print(f"warning: field {h} is not present in this dataset")
        else:
            f = datastore.get_reader(src_assessments[h])
            combined_results = combined_results and f[:]

    inconsistent_healthy =\
        datastore.get_numeric_writer(dest_assessments,
                                     'inconsistent_healthy', 'bool', timestamp)
    inconsistent_not_healthy = \
        datastore.get_numeric_writer(dest_assessments,
                                     'inconsistent_not_healthy', 'bool', timestamp)

    inconsistent_healthy.write((health_status_array == 1) & (combined_results is False))
    inconsistent_not_healthy.write((health_status_array == 2) & (combined_results is True))



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hystore.core.utils import check_input_lengths
from hystore.core import persistence as persist


class CalculateAgeFromYearOfBirth:

    def __init__(self, f_missing_age, f_bad_age, in_range_fn, current_year):
        self.f_missing_age = f_missing_age
        self.f_bad_age = f_bad_age
        self.in_range_fn = in_range_fn
        self.current_year = current_year

    def __call__(self, year_of_birth, age, flags):
        check_input_lengths(('year_of_birth', 'age'), (year_of_birth, age))

        for i_r in range(len(year_of_birth)):
            yob = year_of_birth[i_r]
            if yob != '':
                a = self.current_year - int(float(yob))
                if not self.in_range_fn(a):
                    flags[i_r] |= self.f_bad_age
                    age[i_r] = 0
                else:
                    age[i_r] = a
            else:
                flags[i_r] |= self.f_missing_age


def calculate_age_from_year_of_birth(destination,
                                     year_of_birth, in_range_fn, current_year,
                                     chunksize=None, timestamp=None, name='age'):
    age_18_to_90 = persist.NumericWriter(
        destination, chunksize, '18_to_90', timestamp, 'uint8',
        needs_filter=True)
    age = persist.NumericWriter(
        destination, chunksize, 'age', timestamp, 'int32')
    for y in persist.IndexedStringReader(year_of_birth):
        try:
            a = current_year - int(float(y))
            age_18_to_90.append(in_range_fn(a))

        except ValueError:
            age_18_to_90.append(0)
            age.append(None)


def calculate_age_from_year_of_birth_fast(datastore, min_age, max_age,
                                          year_of_birth, year_of_birth_filter,
                                          age, age_filter, age_range_filter, year,
                                          chunksize=None, timestamp=None):
    yob_v = datastore.get_reader(year_of_birth)
    yob_f = datastore.get_reader(year_of_birth_filter)
    raw_ages = year - yob_v[:]
    raw_age_filter = yob_f[:]
    raw_age_range_filter = raw_age_filter & (min_age <= raw_ages) & (raw_ages <= max_age)
    age.write_part(raw_ages)
    age_filter.write_part(raw_age_filter)
    age_range_filter.write_part(raw_age_range_filter)
    age.flush()
    age_filter.flush()
    age_range_filter.flush()


    # length = len(yob_v)
    # chunksize = age.chunksize if chunksize is None else chunksize
    # for c in range(math.ceil(len(yob_v)) / chunksize):
    #     src_index_start = c * chunksize
    #     maxindex =\
    #         chunksize if (c+1) * chunksize <= length else length % chunksize
    #     src_index_end = src_index_start + maxindex
    #
    #     src_yob_vals = yob_v[src_index_start:src_index_end]
    #     src_yob_flts = yob_f[src_index_start:src_index_end]
    #     raw_ages = src_yob_vals - year
    #     age.values[:maxindex] = raw_ages
    #     age_filter.values[:maxindex] =\
    #         np.logical_not(src_yob_flts) & (min_age <= raw_ages) & (raw_ages <= max_age)
    #     age.write_chunk(maxindex)
    #     age_filter.write_chunk(maxindex)
    #
    # age.flush()
    # age_filter.flush()



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from hystore.core.utils import check_input_lengths


class CheckTestingConsistency:
    def __init__(self, f_inconsistent_not_tested, f_inconsistent_tested):
        self.f_inconsistent_not_tested = f_inconsistent_not_tested
        self.f_inconsistent_tested = f_inconsistent_tested

    def __call__(self, had_test, test_result, flags):
        # if len(had_test) != len(test_result):
        #     error_str = "'had_test' (length {} must be the same length as 'test_result' (length {})"
        #     raise ValueError(error_str.format(len(had_test), len(test_result)))
        # if len(had_test) != len(flags):
        #     error_str = "'had_test' (length {} must be the same length as 'flags' (length {})"
        #     raise ValueError(error_str.format(len(had_test), len(flags)))
        check_input_lengths(('had_test', 'test_result', 'flags'), (had_test, test_result, flags))

        for i_r in range(len(had_test)):
            ht = had_test[i_r]
            tr = test_result[i_r]
            if ht != 2 and tr != 0:
                flags[i_r] |= self.f_inconsistent_not_tested
            if ht == 2 and tr == 0:
                flags[i_r] |= self.f_inconsistent_tested



import numpy as np

dict_coef = {
    'intercept': np.float64(-1.32167356),
    'persistent_cough': np.float64(0.30872189),
    'skipped_meals': np.float64(0.38691672),
    'loss_of_smell': np.float64(1.74711853),
    'gender': np.float64(0.43576578),
    'age': np.float64(-0.00716455),
    'fatigue': np.float64(0.49534113)
}
dict_sd = {
    'intercept': np.float64(0.080211442),
    'persistent_cough': np.float64(0.041716425),
    'skipped_meals': np.float64(0.046133888),
    'loss_of_smell': np.float64(0.042320229),
    'gender': np.float64(0.047851708),
    'age': np.float64(0.001729339),
    'fatigue': np.float64(0.052359487)
}


def logit_prediction(df, dict_coef, index):
    temp_sum = np.float64(0)
    for k in dict_coef.keys():
        if k != 'intercept':
            data = df[k]
            # print(f"df[{k}] range: min={df[k].min()}, max={df[k].max()}")
            # print(f"type of dict_coef item: {type(dict_coef[k][index])}, {dict_coef[k][index]}")
            temp_sum += data * dict_coef[k][index]
    temp_sum += dict_coef['intercept'][index]
    # print(f"temp_sum type: {type(temp_sum)}")
    #temp_logit = 1.0/(1+np.exp(-1.0*temp_sum.astype(float)))
    # print(f"temp_sum min={temp_sum.min()}, max={temp_sum.max()}")
    temp_logit = 1.0 / (1 + np.exp(-1.0 * temp_sum))
    return temp_logit


def imputation_pos_all(df_umerge, distr, samplecount, threshold=None):
    # df_umerge = pd.merge(df_assess, df_patred, left_on='patient_id', right_on='id', how='left')
    # df_umerge = df_umerge.drop_duplicates(['patient_id', 'interval_days']) #not needed due to daily assessments
    list_temp = []

    df_umerge = {k: v[:] for k, v in df_umerge.items()}

    results = logit_prediction(df_umerge, distr, 0)
    for i in range(1, samplecount):
        print(f"  imputation iteration {i}")
        results += logit_prediction(df_umerge, distr, i)
    average_decision = results / np.float64(samplecount)
    # average_decision = np.mean(np.asarray(list_temp).astype(float), 0)
    # this could be a threshold of
    # print(average_decision.shape, len(average_decision > threshold), df_umerge.shape, df_assess.shape)
    # df_assess['imputed'] = np.where(average_decision > threshold, 1, 0) # this goes to a writer
    # return df_assess
    if threshold is not None:
        return np.where(average_decision > threshold, np.uint8(1), np.uint8(0))
    else:
        return average_decision


def nature_medicine_model_1(persistent_cough, skipped_meals, loss_of_smell, fatigue,
                            age_in_assessment_space, gender_in_assessment_space,
                            threshold=None):
    dataset = {
        'persistent_cough': persistent_cough,
        'skipped_meals': skipped_meals,
        'loss_of_smell': loss_of_smell,
        'fatigue': fatigue,
        'age': age_in_assessment_space,
        'gender': gender_in_assessment_space
    }

    samplecount = 100
    distr = {}
    for k in dict_coef.keys():
        distr[k] = dict_coef[k] + dict_sd[k] * np.random.randn(samplecount)
    # patients (age , gender)

    results = imputation_pos_all(dataset, distr, samplecount, threshold)
    return results



# Copyright 2020 KCL-BMEIS - King's College London
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import time
from collections import defaultdict

import numpy as np
import numba

from hystore.processing.age_from_year_of_birth import calculate_age_from_year_of_birth_fast
from hystore.processing.weight_height_bmi import weight_height_bmi_fast_1
from hystore.processing.inconsistent_symptoms import check_inconsistent_symptoms_1
from hystore.processing.temperature import validate_temperature_1
from hystore.processing.combined_healthcare_worker import combined_hcw_with_contact
from hystore.core import persistence
from hystore.core.persistence import DataStore
from hystore.core import readerwriter as rw


# TODO: hard filter
# TODO: journalling for hdf5 robustness

# TODO: postprocessing activities
# * assessment sort by (patient_id, created_at)
# * aggregate from assessments to patients
#   * was first unwell
#   * first assessment
#   * last assessment
#   * assessment count
#   * assessment index start
#   * assessment index end

def log(*a, **kwa):
    print(*a, **kwa)


def postprocess(dataset, destination, timestamp=None, flags='all'):

    ds = DataStore(timestamp=timestamp)
    patients_src = dataset['patients']
    patients_dest = ds.get_or_create_group(destination, 'patients')
    assessments_src = dataset['assessments']
    assessments_dest = ds.get_or_create_group(destination, 'assessments')
    daily_assessments_dest = ds.get_or_create_group(destination, 'daily_assessments')
    tests_src = dataset['tests']
    tests_dest = ds.get_or_create_group(destination, 'tests')

    sort_enabled = lambda x: True
    process_enabled = lambda x: True

    sort_patients = sort_enabled(flags) and True
    sort_assessments = sort_enabled(flags) and True
    sort_tests = sort_enabled(flags) and True

    make_assessment_patient_id_fkey = process_enabled(flags) and True
    year_from_age = process_enabled(flags) and True
    clean_weight_height_bmi = process_enabled(flags) and True
    health_worker_with_contact = process_enabled(flags) and True
    clean_temperatures = process_enabled(flags) and True
    check_symptoms = process_enabled(flags) and True
    create_daily = process_enabled(flags) and True
    make_patient_level_assessment_metrics = process_enabled(flags) and True
    make_patient_level_daily_assessment_metrics = process_enabled(flags) and True
    make_new_test_level_metrics = process_enabled(flags) and True

    # post process patients
    # TODO: need an transaction table

    print(patients_src.keys())
    print(dataset['assessments'].keys())
    print(dataset['tests'].keys())

    # write_mode = 'overwrite'
    write_mode = 'write'

    # Sorting
    # #######

    if sort_patients:
        duplicate_filter =\
            persistence.filter_duplicate_fields(ds.get_reader(patients_src['id'])[:])
        print("duplicate True/False:", duplicate_filter)

        for k in patients_src.keys():
            t0 = time.time()
            r = ds.get_reader(patients_src[k])
            w = r.get_writer(patients_dest, k)
            ds.apply_filter(duplicate_filter, r, w)
            print(f"'{k}' filtered in {time.time() - t0}s")

        print(np.count_nonzero(duplicate_filter == True),
              np.count_nonzero(duplicate_filter == False))
        sort_keys = ('id',)
        ds.sort_on(
            patients_dest, patients_dest, sort_keys, write_mode='overwrite')

    if sort_assessments:
        sort_keys = ('patient_id', 'created_at')
        ds.sort_on(
            assessments_src, assessments_dest, sort_keys)

    if sort_tests:
        sort_keys = ('patient_id', 'created_at')
        ds.sort_on(tests_src, tests_dest, sort_keys)


    # Processing
    # ##########

    # sorted_patients_src = patients_dest if sort_patients else patients_src
    # sorted_assessments_src = assessments_dest if sort_assessments else assessments_src
    sorted_patients_src = patients_dest
    sorted_assessments_src = assessments_dest

    # Patient processing
    # ==================

    if year_from_age:
        log("year of birth -> age; 18 to 90 filter")
        t0 = time.time()
        age = ds.get_numeric_writer(patients_dest, 'age', 'uint32',
                                    write_mode)
        age_filter = ds.get_numeric_writer(patients_dest, 'age_filter',
                                           'bool', write_mode)
        age_16_to_90 = ds.get_numeric_writer(patients_dest, '16_to_90_years',
                                             'bool', write_mode)
        print('year_of_birth:', patients_dest['year_of_birth'])
        for k in patients_dest['year_of_birth'].attrs.keys():
            print(k, patients_dest['year_of_birth'].attrs[k])
        calculate_age_from_year_of_birth_fast(
            ds, 16, 90,
            patients_dest['year_of_birth'], patients_dest['year_of_birth_valid'],
            age, age_filter, age_16_to_90,
            2020)
        log(f"completed in {time.time() - t0}")

        print('age_filter count:', np.sum(patients_dest['age_filter']['values'][:]))
        print('16_to_90_years count:', np.sum(patients_dest['16_to_90_years']['values'][:]))

    if clean_weight_height_bmi:
        log("height / weight / bmi; standard range filters")
        t0 = time.time()

        weights_clean = ds.get_numeric_writer(patients_dest, 'weight_kg_clean',
                                              'float32', write_mode)
        weights_filter = ds.get_numeric_writer(patients_dest, '40_to_200_kg',
                                               'bool', write_mode)
        heights_clean = ds.get_numeric_writer(patients_dest, 'height_cm_clean',
                                              'float32', write_mode)
        heights_filter = ds.get_numeric_writer(patients_dest, '110_to_220_cm',
                                               'bool', write_mode)
        bmis_clean = ds.get_numeric_writer(patients_dest, 'bmi_clean',
                                           'float32', write_mode)
        bmis_filter = ds.get_numeric_writer(patients_dest, '15_to_55_bmi',
                                            'bool', write_mode)

        weight_height_bmi_fast_1(ds, 40, 200, 110, 220, 15, 55,
                                 None, None, None, None,
                                 patients_dest['weight_kg'], patients_dest['weight_kg_valid'],
                                 patients_dest['height_cm'], patients_dest['height_cm_valid'],
                                 patients_dest['bmi'], patients_dest['bmi_valid'],
                                 weights_clean, weights_filter, None,
                                 heights_clean, heights_filter, None,
                                 bmis_clean, bmis_filter, None)
        log(f"completed in {time.time() - t0}")

    if health_worker_with_contact:
        log("health_worker_with_contact field")
        #writer = ds.get_categorical_writer(patients_dest, 'health_worker_with_contact', 'int8')
        combined_hcw_with_contact(ds,
                                  ds.get_reader(patients_dest['healthcare_professional']),
                                  ds.get_reader(patients_dest['contact_health_worker']),
                                  ds.get_reader(patients_dest['is_carer_for_community']),
                                  patients_dest, 'health_worker_with_contact')

    # Assessment processing
    # =====================

    if make_assessment_patient_id_fkey:
        print("creating 'assessment_patient_id_fkey' foreign key index for 'patient_id'")
        t0 = time.time()
        patient_ids = ds.get_reader(sorted_patients_src['id'])
        assessment_patient_ids =\
            ds.get_reader(sorted_assessments_src['patient_id'])
        assessment_patient_id_fkey =\
            ds.get_numeric_writer(assessments_dest, 'assessment_patient_id_fkey', 'int64')
        ds.get_index(patient_ids, assessment_patient_ids, assessment_patient_id_fkey)
        print(f"completed in {time.time() - t0}s")


    if clean_temperatures:
        print("clean temperatures")
        t0 = time.time()
        temps = ds.get_reader(sorted_assessments_src['temperature'])
        temp_units = ds.get_reader(sorted_assessments_src['temperature_unit'])
        temps_valid = ds.get_reader(sorted_assessments_src['temperature_valid'])
        dest_temps = temps.get_writer(assessments_dest, 'temperature_c_clean', write_mode)
        dest_temps_valid =\
            temps_valid.get_writer(assessments_dest, 'temperature_35_to_42_inclusive', write_mode)
        dest_temps_modified =\
            temps_valid.get_writer(assessments_dest, 'temperature_modified', write_mode)
        validate_temperature_1(35.0, 42.0,
                               temps, temp_units, temps_valid,
                               dest_temps, dest_temps_valid, dest_temps_modified)
        print(f"temperature cleaning done in {time.time() - t0}")


    if check_symptoms:
        print('check inconsistent health_status')
        t0 = time.time()
        check_inconsistent_symptoms_1(ds, sorted_assessments_src, assessments_dest)
        print(time.time() - t0)


    # Test processing
    # ===============


    # Daily assessments
    # =================

    if create_daily:
        print("generate daily assessments")
        patient_ids = ds.get_reader(sorted_assessments_src['patient_id'])
        created_at_days = ds.get_reader(sorted_assessments_src['created_at_day'])
        raw_created_at_days = created_at_days[:]

        if 'assessment_patient_id_fkey' in assessments_src.keys():
            patient_id_index = assessments_src['assessment_patient_id_fkey']
        else:
            patient_id_index = assessments_dest['assessment_patient_id_fkey']
        patient_id_indices = ds.get_reader(patient_id_index)
        raw_patient_id_indices = patient_id_indices[:]


        print("Calculating patient id index spans")
        t0 = time.time()
        patient_id_index_spans = ds.get_spans(fields=(raw_patient_id_indices,
                                                     raw_created_at_days))
        print(f"Calculated {len(patient_id_index_spans)-1} spans in {time.time() - t0}s")


        print("Applying spans to 'health_status'")
        t0 = time.time()
        default_behavour_overrides = {
            'id': ds.apply_spans_last,
            'patient_id': ds.apply_spans_last,
            'patient_index': ds.apply_spans_last,
            'created_at': ds.apply_spans_last,
            'created_at_day': ds.apply_spans_last,
            'updated_at': ds.apply_spans_last,
            'updated_at_day': ds.apply_spans_last,
            'version': ds.apply_spans_max,
            'country_code': ds.apply_spans_first,
            'date_test_occurred': None,
            'date_test_occurred_guess': None,
            'date_test_occurred_day': None,
            'date_test_occurred_set': None,
        }
        for k in sorted_assessments_src.keys():
            t1 = time.time()
            reader = ds.get_reader(sorted_assessments_src[k])
            if k in default_behavour_overrides:
                apply_span_fn = default_behavour_overrides[k]
                if apply_span_fn is not None:
                    apply_span_fn(patient_id_index_spans, reader,
                                  reader.get_writer(daily_assessments_dest, k))
                    print(f"  Field {k} aggregated in {time.time() - t1}s")
                else:
                    print(f"  Skipping field {k}")
            else:
                if isinstance(reader, rw.CategoricalReader):
                    ds.apply_spans_max(
                        patient_id_index_spans, reader,
                        reader.get_writer(daily_assessments_dest, k))
                    print(f"  Field {k} aggregated in {time.time() - t1}s")
                elif isinstance(reader, rw.IndexedStringReader):
                    ds.apply_spans_concat(
                        patient_id_index_spans, reader,
                        reader.get_writer(daily_assessments_dest, k))
                    print(f"  Field {k} aggregated in {time.time() - t1}s")
                elif isinstance(reader, rw.NumericReader):
                    ds.apply_spans_max(
                        patient_id_index_spans, reader,
                        reader.get_writer(daily_assessments_dest, k))
                    print(f"  Field {k} aggregated in {time.time() - t1}s")
                else:
                    print(f"  No function for {k}")

        print(f"apply_spans completed in {time.time() - t0}s")


    # TODO - patient measure: assessments per patient

    if make_patient_level_assessment_metrics:
        if 'assessment_patient_id_fkey' in assessments_dest:
            src = assessments_dest['assessment_patient_id_fkey']
        else:
            src = assessments_src['assessment_patient_id_fkey']
        assessment_patient_id_fkey = ds.get_reader(src)
        # generate spans from the assessment-space patient_id foreign key
        spans = ds.get_spans(field=assessment_patient_id_fkey)

        ids = ds.get_reader(patients_dest['id'])

        # print('predicate_and_join')
        # acpp2 = ds.get_numeric_writer(patients_dest, 'assessment_count_2', 'uint32')
        # ds.predicate_and_join(ds.apply_spans_count, ids,
        #                              assessment_patient_id_fkey, None, acpp2, spans)

        print('calculate assessment counts per patient')
        t0 = time.time()
        writer = ds.get_numeric_writer(patients_dest, 'assessment_count', 'uint32')
        aggregated_counts = ds.aggregate_count(fkey_index_spans=spans)
        ds.join(ids, assessment_patient_id_fkey, aggregated_counts, writer, spans)
        print(f"calculated assessment counts per patient in {time.time() - t0}")

        print('calculate first assessment days per patient')
        t0 = time.time()
        reader = ds.get_reader(sorted_assessments_src['created_at_day'])
        writer = ds.get_fixed_string_writer(patients_dest, 'first_assessment_day', 10)
        aggregated_counts = ds.aggregate_first(fkey_index_spans=spans, reader=reader)
        ds.join(ids, assessment_patient_id_fkey, aggregated_counts, writer, spans)
        print(f"calculated first assessment days per patient in {time.time() - t0}")

        print('calculate last assessment days per patient')
        t0 = time.time()
        reader = ds.get_reader(sorted_assessments_src['created_at_day'])
        writer = ds.get_fixed_string_writer(patients_dest, 'last_assessment_day', 10)
        aggregated_counts = ds.aggregate_last(fkey_index_spans=spans, reader=reader)
        ds.join(ids, assessment_patient_id_fkey, aggregated_counts, writer, spans)
        print(f"calculated last assessment days per patient in {time.time() - t0}")

        print('calculate maximum assessment test result per patient')
        t0 = time.time()
        reader = ds.get_reader(sorted_assessments_src['tested_covid_positive'])
        writer = reader.get_writer(patients_dest, 'max_assessment_test_result')
        max_result_value = ds.aggregate_max(fkey_index_spans=spans, reader=reader)
        ds.join(ids, assessment_patient_id_fkey, max_result_value, writer, spans)
        print(f"calculated maximum assessment test result in {time.time() - t0}")

    # TODO - patient measure: daily assessments per patient

    if make_patient_level_daily_assessment_metrics:
        print("creating 'daily_assessment_patient_id_fkey' foreign key index for 'patient_id'")
        t0 = time.time()
        patient_ids = ds.get_reader(sorted_patients_src['id'])
        daily_assessment_patient_ids =\
            ds.get_reader(daily_assessments_dest['patient_id'])
        daily_assessment_patient_id_fkey =\
            ds.get_numeric_writer(daily_assessments_dest, 'daily_assessment_patient_id_fkey',
                                  'int64')
        ds.get_index(patient_ids, daily_assessment_patient_ids,
                     daily_assessment_patient_id_fkey)
        print(f"completed in {time.time() - t0}s")

        spans = ds.get_spans(
            field=ds.get_reader(daily_assessments_dest['daily_assessment_patient_id_fkey']))

        print('calculate daily assessment counts per patient')
        t0 = time.time()
        writer = ds.get_numeric_writer(patients_dest, 'daily_assessment_count', 'uint32')
        aggregated_counts = ds.aggregate_count(fkey_index_spans=spans)
        daily_assessment_patient_id_fkey =\
            ds.get_reader(daily_assessments_dest['daily_assessment_patient_id_fkey'])
        ds.join(ids, daily_assessment_patient_id_fkey, aggregated_counts, writer, spans)
        print(f"calculated daily assessment counts per patient in {time.time() - t0}")


    # TODO - new test count per patient:
    if make_new_test_level_metrics:
        print("creating 'test_patient_id_fkey' foreign key index for 'patient_id'")
        t0 = time.time()
        patient_ids = ds.get_reader(sorted_patients_src['id'])
        test_patient_ids = ds.get_reader(tests_dest['patient_id'])
        test_patient_id_fkey = ds.get_numeric_writer(tests_dest, 'test_patient_id_fkey',
                                                     'int64')
        ds.get_index(patient_ids, test_patient_ids, test_patient_id_fkey)
        test_patient_id_fkey = ds.get_reader(tests_dest['test_patient_id_fkey'])
        spans = ds.get_spans(field=test_patient_id_fkey)
        print(f"completed in {time.time() - t0}s")

        print('calculate test_counts per patient')
        t0 = time.time()
        writer = ds.get_numeric_writer(patients_dest, 'test_count', 'uint32')
        aggregated_counts = ds.aggregate_count(fkey_index_spans=spans)
        ds.join(ids, test_patient_id_fkey, aggregated_counts, writer, spans)
        print(f"calculated test counts per patient in {time.time() - t0}")

        print('calculate test_result per patient')
        t0 = time.time()
        test_results = ds.get_reader(tests_dest['result'])
        writer = test_results.get_writer(patients_dest, 'max_test_result')
        aggregated_results = ds.aggregate_max(fkey_index_spans=spans, reader=test_results)
        ds.join(ids, test_patient_id_fkey, aggregated_results, writer, spans)
        print(f"calculated max_test_result per patient in {time.time() - t0}")
